"""
A module containing various modulation-related plotting functions.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from typing_extensions import Literal

from .._helper import export, verify_arraylike, verify_bool, verify_isinstance, verify_scalar
from .._link_budget._capacity import shannon_limit_ebn0 as _shannon_limit_ebn0
from .._modulation import LinearModulation, PiMPSK
from ._helper import verify_sample_rate
from ._rc_params import RC_PARAMS
from ._time_domain import raster


@export
def constellation(
    x_hat: npt.ArrayLike,
    limits: tuple[float, float] | None = None,
    persistence: bool = False,
    colorbar: bool = True,
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Plots the constellation of the complex symbols $\hat{x}[k]$.

    Arguments:
        x_hat: The complex symbols $\hat{x}[k]$.
        limits: The axis limits, which apply to both the x- and y-axis. If `None`, the axis limits are
            set to 10% larger than the maximum value.
        persistence: Indicates whether to plot the points as a persistence plot. A persistence plot is a
            2D histogram of the points.
        colorbar: Indicates whether to add a colorbar to the plot. This is only added if `persistence=True`.
        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to Matplotlib functions.

            If `persistence=False`, the following keyword arguments are passed to :func:`matplotlib.pyplot.scatter`.
            The defaults may be overwritten.

            - `"marker"`: `"."`
            - `"linestyle"`: `"none"`

            If `persistence=True`, the following keyword arguments are passed to :func:`numpy.histogram2d` and
            :func:`matplotlib.pyplot.pcolormesh`. The defaults may be overwritten.

            - `"range"`: +/- 10% of the maximum value
            - `"bins"`: `100  # Number of bins per axis`
            - `"cmap"`: `"rainbow"`
            - `"show_zero"`: `False`

    Example:
        Display the symbol constellation for Gray-coded QPSK at 6 dB $E_s/N_0$.

        .. ipython:: python

            qpsk = sdr.PSK(4, phase_offset=45); \
            s = np.random.randint(0, qpsk.order, 100_000); \
            x = qpsk.map_symbols(s); \
            x_hat = sdr.awgn(x, 6);

            @savefig sdr_plot_constellation_1.png
            plt.figure(); \
            sdr.plot.constellation(x_hat[0:1_000])

        Display the symbol constellation using a persistence plot.

        .. ipython:: python

            @savefig sdr_plot_constellation_2.png
            plt.figure(); \
            sdr.plot.constellation(x_hat, persistence=True)

    Group:
        plot-modulation
    """
    x_hat = verify_arraylike(x_hat, complex=True, ndim=1)
    verify_bool(persistence)
    verify_bool(colorbar)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        # Set the axis limits to 10% larger than the maximum value
        if limits is None:
            lim = np.max(np.abs(x_hat)) * 1.1
            lim = max(lim, ax.get_ylim()[1])  # Don't reduce the limits if they are already larger
            limits = (-lim, lim)

        if persistence:
            default_kwargs = {
                "range": (limits, limits),
                "bins": 100,  # Number of bins per axis
                "cmap": "rainbow",
                "show_zero": False,
            }
            kwargs = {**default_kwargs, **kwargs}

            bins = kwargs.pop("bins")
            range = kwargs.pop("range")
            h, t_edges, x_edges = np.histogram2d(x_hat.real, x_hat.imag, bins=bins, range=range)

            cmap = kwargs.pop("cmap")  # Need to pop cmap to avoid passing it twice to pcolormesh
            cmap = plt.colormaps[cmap]
            show_zero = kwargs.pop("show_zero")
            if show_zero:
                cmap = cmap.with_extremes(bad=cmap(0))
            else:
                h[h == 0] = np.nan  # Set 0s to NaNs so they don't show up in the plot

            pcm = ax.pcolormesh(t_edges, x_edges, h.T, cmap=cmap, **kwargs)
            if colorbar:
                plt.colorbar(pcm, label="Points", pad=0.05)
        else:
            default_kwargs = {
                "marker": ".",
                "linestyle": "none",
            }
            kwargs = {**default_kwargs, **kwargs}
            ax.plot(x_hat.real, x_hat.imag, **kwargs)

        ax.axis("square")
        ax.set_xlim(limits)
        ax.set_ylim(limits)
        if "label" in kwargs:
            ax.legend()
        ax.set_xlabel("In-phase channel, $I$")
        ax.set_ylabel("Quadrature channel, $Q$")
        ax.set_title("Constellation")


@export
def symbol_map(
    modulation: LinearModulation | npt.ArrayLike,
    annotate: bool | Literal["bin"] = True,
    limits: tuple[float, float] | None = None,
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Plots the symbol map of the complex symbols $\hat{x}[k]$.

    Arguments:
        modulation: The linear modulation or symbol map $\{0, \dots, M-1\} \mapsto \mathbb{C}$.
        annotate: If `True`, the symbols are annotated with their index.
            If `"bin"`, the symbols are annotated with their binary representation.
        limits: The axis limits, which apply to both the x- and y-axis.
            If `None`, the axis limits are set to 50% larger than the maximum value.
        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.
            The following keyword arguments are set by default. The defaults may be overwritten.

            - `"marker"`: `"x"`
            - `"markersize"`: 6
            - `"linestyle"`: `"none"`

    Example:
        Display the symbol mapping for Gray-coded QPSK.

        .. ipython:: python

            qpsk = sdr.PSK(4, phase_offset=45)

            @savefig sdr_plot_symbol_map_1.png
            plt.figure(); \
            sdr.plot.symbol_map(qpsk.symbol_map)

    Group:
        plot-modulation
    """
    verify_isinstance(modulation, (LinearModulation, np.ndarray))
    verify_isinstance(annotate, (bool, str))
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if isinstance(modulation, LinearModulation):
            symbol_map_ = modulation.symbol_map
        else:
            symbol_map_ = np.asarray(modulation)
        k = int(np.log2(symbol_map_.size))

        if isinstance(modulation, PiMPSK):
            label = kwargs.pop("label", None)
            if label:
                even_label = f"{label} (even)"
                odd_label = f"{label} (odd)"
            else:
                even_label = "even"
                odd_label = "odd"
            even_symbol_map = symbol_map_
            odd_symbol_map = symbol_map_ * np.exp(1j * np.pi / modulation.order)
            symbol_map(even_symbol_map, annotate=annotate, limits=limits, label=even_label, **kwargs)
            symbol_map(odd_symbol_map, annotate=annotate, limits=limits, label=odd_label, **kwargs)
            return

        # Set the axis limits to 50% larger than the maximum value
        if limits is None:
            lim = np.max(np.abs(symbol_map_)) * 1.5
            limits = (-lim, lim)

        default_kwargs = {
            "marker": "x",
            "markersize": 6,
            "linestyle": "none",
        }
        kwargs = {**default_kwargs, **kwargs}
        ax.plot(symbol_map_.real, symbol_map_.imag, **kwargs)

        if annotate:
            for i, symbol in enumerate(symbol_map_):
                if annotate == "bin":
                    label = f"{i} = " + np.binary_repr(i, k)
                else:
                    label = i

                ax.annotate(
                    label,
                    (symbol.real, symbol.imag),
                    xytext=(0, 5),
                    textcoords="offset points",
                    ha="center",
                    va="bottom",
                    fontsize=8,
                )

        ax.axis("square")
        ax.set_xlim(limits)
        ax.set_ylim(limits)
        if "label" in kwargs:
            ax.legend()
        ax.set_xlabel("In-phase channel, $I$")
        ax.set_ylabel("Quadrature channel, $Q$")
        ax.set_title("Symbol map")


@export
def eye(
    x: npt.ArrayLike,
    sps: int,
    span: int = 2,
    sample_rate: float | None = None,
    color: Literal["index"] | str = "index",
    persistence: bool = False,
    colorbar: bool = True,
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Plots the eye diagram of the baseband modulated signal $x[n]$.

    Arguments:
        x: The baseband modulated signal $x[n]$. If `x` is complex, in-phase and quadrature eye diagrams are plotted
            in separate subplots.
        sps: The number of samples per symbol.
        span: The number of symbols per raster.
        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Symbol".
        color: Indicates how to color the rasters. If `"index"`, the rasters are colored based on their index.
            If a valid Matplotlib color, the rasters are all colored with that color.
        persistence: Indicates whether to plot the raster as a persistence plot. A persistence plot is a
            2D histogram of the rasters.
        colorbar: Indicates whether to add a colorbar to the plot. This is only added if `color="index"` or
            `persistence=True`.
        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to :func:`sdr.plot.raster`.

    Example:
        Modulate 1,000 QPSK symbols using a square root raised cosine (SRRC) pulse shaping filter. The SRRC pulse shape
        is not a Nyquist filter, and intersymbol interference (ISI) can be observed in the eye diagram. Plot the eye
        diagram using index-colored rasters. Note, we are ignoring the transient response of the pulse shaping filter
        at the beginning and end of the signal.

        .. ipython:: python

            psk = sdr.PSK(4, phase_offset=45, pulse_shape="srrc"); \
            sps = psk.sps; \
            s = np.random.randint(0, psk.order, 1_000); \
            tx_samples = psk.modulate(s)

            @savefig sdr_plot_eye_1.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(tx_samples[4*sps : -4*sps], sps); \
            plt.suptitle("Transmitted QPSK symbols with SRRC pulse shape");

        Plot the eye diagram using a persistence plot. This provides insight into the probability density
        function of the signal.

        .. ipython:: python

            @savefig sdr_plot_eye_2.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(tx_samples[4*sps : -4*sps], sps, persistence=True); \
            plt.suptitle("Transmitted QPSK symbols with SRRC pulse shape");

        Apply a SRRC matched filter at the receiver. The cascaded transmit and receive SRRC filters are equivalent
        to a single raised cosine (RC) filter, which is a Nyquist filter. ISI is no longer observed, and the eye is
        open.

        .. ipython:: python

            mf = sdr.FIR(psk.pulse_shape); \
            rx_samples = mf(tx_samples, mode="same")

            @savefig sdr_plot_eye_3.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(rx_samples[4*sps : -4*sps], sps, persistence=True); \
            plt.suptitle("Received and matched filtered QPSK symbols");

    Group:
        plot-modulation
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    verify_scalar(sps, int=True, inclusive_min=2)
    verify_scalar(span, int=True, positive=True)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate, default=sps)
    verify_isinstance(color, str)
    verify_bool(persistence)
    verify_bool(colorbar)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):

        def _eye(ax, xx):
            raster(
                xx,
                length=span * sps + 1,
                stride=sps,
                sample_rate=sample_rate,
                color=color,
                persistence=persistence,
                colorbar=colorbar,
                ax=ax,
                **kwargs,
            )
            if not sample_rate_provided:
                ax.set_xlabel("Symbol, $k$")

            # Make y-axis symmetric
            ymin, ymax = ax.get_ylim()
            ylim = max(np.abs(ymin), np.abs(ymax))
            ax.set_ylim(-ylim, ylim)

        if np.iscomplexobj(x):
            ax = plt.subplot(2, 1, 1)
            _eye(ax, x.real)
            plt.title("In-phase eye diagram")
            ax = plt.subplot(2, 1, 2)
            _eye(ax, x.imag)
            plt.title("Quadrature eye diagram")
        else:
            if ax is None:
                ax = plt.gca()
            _eye(ax, x)
            plt.title("Eye diagram")


@export
def phase_tree(
    x: npt.ArrayLike,
    sps: int,
    span: int = 2,
    sample_rate: float | None = None,
    color: Literal["index"] | str = "index",
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Plots the phase tree of a continuous-phase modulated (CPM) signal signal $x[n]$.

    Arguments:
        x: The baseband CPM signal $x[n]$.
        sps: The number of samples per symbol.
        span: The number of symbols per raster.
        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Symbol".
        color: Indicates how to color the rasters. If `"index"`, the rasters are colored based on their index.
            If a valid Matplotlib color, the rasters are all colored with that color.
        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to :func:`sdr.plot.raster`.

    Example:
        Modulate 100 MSK symbols.

        .. ipython:: python

            msk = sdr.MSK(); \
            s = np.random.randint(0, msk.order, 100); \
            x = msk.modulate(s)

        .. ipython:: python

            @savefig sdr_plot_phase_tree_1.png
            plt.figure(); \
            sdr.plot.phase_tree(x, msk.sps)

    Group:
        plot-modulation
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    verify_scalar(sps, int=True, inclusive_min=2)
    verify_scalar(span, int=True, positive=True)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate, default=sps)
    verify_isinstance(color, str)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        phase = np.angle(x)

        # Create a strided array of phase values
        length = sps * span + 1
        stride = sps
        N_rasters = (phase.size - length) // stride + 1
        phase_strided = np.lib.stride_tricks.as_strided(
            phase, shape=(N_rasters, length), strides=(phase.strides[0] * stride, phase.strides[0]), writeable=False
        )

        # Unwrap the phase and convert to degrees
        phase_strided = np.unwrap(phase_strided, axis=1)
        phase_strided -= phase_strided[:, 0][:, np.newaxis]  # Normalize to 0 degrees at the first sample
        phase_strided = np.rad2deg(phase_strided)

        raster(
            phase_strided,
            sample_rate=sample_rate,
            color=color,
            ax=ax,
            **kwargs,
        )
        if not sample_rate_provided:
            ax.set_xlabel("Symbol, $k$")

        # Make y-axis symmetric and have ticks every 180 degrees
        ymin, ymax = ax.get_ylim()
        ylim = max(np.abs(ymin), np.abs(ymax))
        ylim = np.ceil(ylim / 180) * 180
        ax.set_ylim(-ylim, ylim)
        ax.set_yticks(np.arange(-ylim, ylim + 1, 180))

        ax.set_ylabel(r"Phase (deg), $\phi$")


@export
def ber(
    ebn0: npt.ArrayLike,
    ber: npt.ArrayLike,
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Plots the bit error rate (BER) as a function of $E_b/N_0$.

    Arguments:
        ebn0: The bit energy $E_b$ to noise PSD $N_0$ ratio (dB).
        ber: The bit error rate $P_{be}$.
        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.semilogy`.

    Examples:
        Plot theoretical BER curves for BPSK, QPSK, 8-PSK, and 16-PSK in an AWGN channel.

        .. ipython:: python

            bpsk = sdr.PSK(2); \
            qpsk = sdr.PSK(4); \
            psk8 = sdr.PSK(8); \
            psk16 = sdr.PSK(16); \
            ebn0 = np.linspace(-2, 10, 100)

            @savefig sdr_plot_ber_1.png
            plt.figure(); \
            sdr.plot.ber(ebn0, bpsk.ber(ebn0), label="BPSK"); \
            sdr.plot.ber(ebn0, qpsk.ber(ebn0), label="QPSK"); \
            sdr.plot.ber(ebn0, psk8.ber(ebn0), label="8-PSK"); \
            sdr.plot.ber(ebn0, psk16.ber(ebn0), label="16-PSK"); \
            plt.title("BER curves for PSK modulation in an AWGN channel");

    Group:
        plot-modulation
    """
    ebn0 = verify_arraylike(ebn0, float=True, ndim=1)
    ber = verify_arraylike(ber, float=True, ndim=1)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        default_kwargs = {}
        kwargs = {**default_kwargs, **kwargs}

        ax.semilogy(ebn0, ber, **kwargs)
        if "label" in kwargs:
            ax.legend()

        ax.set_xlabel("Bit energy-to-noise PSD ratio (dB), $E_b/N_0$")
        ax.set_ylabel("Probability of bit error, $P_{be}$")
        ax.set_title("Bit error rate curve")


@export
def ser(
    esn0: npt.ArrayLike,
    ser: npt.ArrayLike,
    ax: plt.Axes | None = None,
    **kwargs,
):
    r"""
    Plots the symbol error rate (SER) as a function of $E_s/N_0$.

    Arguments:
        esn0: The symbol energy $E_s$ to noise PSD $N_0$ ratio (dB).
        ser: The symbol error rate $P_{se}$.
        ax: The axis to plot on. If `None`, the current axis is used.
        kwargs: Additional keyword arguments to pass to :func:`matplotlib.pyplot.semilogy`.

    Examples:
        Plot theoretical SER curves for BPSK, QPSK, 8-PSK, and 16-PSK in an AWGN channel.

        .. ipython:: python

            bpsk = sdr.PSK(2); \
            qpsk = sdr.PSK(4); \
            psk8 = sdr.PSK(8); \
            psk16 = sdr.PSK(16); \
            esn0 = np.linspace(-2, 10, 100)

            @savefig sdr_psk_ser_1.png
            plt.figure(); \
            sdr.plot.ser(esn0, bpsk.ser(esn0), label="BPSK"); \
            sdr.plot.ser(esn0, qpsk.ser(esn0), label="QPSK"); \
            sdr.plot.ser(esn0, psk8.ser(esn0), label="8-PSK"); \
            sdr.plot.ser(esn0, psk16.ser(esn0), label="16-PSK"); \
            plt.title("SER curves for PSK modulation in an AWGN channel");

    Group:
        plot-modulation
    """
    esn0 = verify_arraylike(esn0, float=True, ndim=1)
    ser = verify_arraylike(ser, float=True, ndim=1)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        default_kwargs = {}
        kwargs = {**default_kwargs, **kwargs}

        ax.semilogy(esn0, ser, **kwargs)
        if "label" in kwargs:
            ax.legend()

        ax.set_xlabel("Symbol energy-to-noise PSD ratio (dB), $E_s/N_0$")
        ax.set_ylabel("Probability of symbol error, $P_{se}$")
        ax.set_title("Symbol error rate curve")


@export
def shannon_limit_ebn0(
    rho: float,
    ax: plt.Axes | None = None,
):
    r"""
    Plots the Shannon limit for the bit energy-to-noise PSD ratio $E_b/N_0$.

    Arguments:
        rho: The nominal spectral efficiency $\rho$ of the modulation in bits/2D.
        ax: The axis to plot on. If `None`, the current axis is used.

    Examples:
        Plot the absolute Shannon limit on $E_b/N_0$ and the Shannon limit for $\rho = 2$ bits/2D. Compare these to the
        theoretical BER curve for BPSK modulation, which has spectral efficiency $\rho = 2$ bits/2D.

        .. ipython:: python

            bpsk = sdr.PSK(2)
            ebn0 = np.linspace(-2, 10, 201)

            @savefig sdr_plot_shannon_limit_ebn0_1.png
            plt.figure(); \
            sdr.plot.ber(ebn0, bpsk.ber(ebn0), label="BPSK theoretical"); \
            plt.ylim(1e-6, 1e0); \
            sdr.plot.shannon_limit_ebn0(0); \
            sdr.plot.shannon_limit_ebn0(2); \
            plt.title("Bit error rate curve for PSK modulation in AWGN");

    Group:
        plot-modulation
    """
    verify_scalar(rho, float=True, non_negative=True)
    verify_isinstance(ax, plt.Axes, optional=True)

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        ebn0 = _shannon_limit_ebn0(rho)

        ax.axvline(x=ebn0, ymin=0, ymax=0.75, color="black", linestyle="--")
        ax.annotate(
            rf"Shannon limit for $\rho = {rho}$",
            xy=(ebn0, ax.get_ylim()[0]),
            xytext=(-1.25, 0.75),
            textcoords="offset fontsize",
            rotation=90,
        )



"""
A module containing estimation algorithms for signal-to-noise ratio (SNR) calculations.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._conversion import db, linear
from .._helper import convert_output, export, verify_arraylike


@export
def composite_snr(snr1: npt.ArrayLike, snr2: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Calculates the signal-to-noise ratio (SNR) of the product of two signals.

    Arguments:
        snr1: The signal-to-noise ratio (SNR) of the first signal $\gamma_1$ in dB.
        snr2: The signal-to-noise ratio (SNR) of the second signal $\gamma_2$ in dB.

    Returns:
        The signal-to-noise ratio (SNR) of the product of the two signals $\gamma$ in dB.

    Notes:
        The effective or composite SNR $\gamma$ of the product of two signals with SNRs $\gamma_1$ and $\gamma_2$
        is given by

        $$\frac{1}{\gamma} = \frac{1}{\gamma_1} + \frac{1}{\gamma_2} + \frac{1}{\gamma_1 \gamma_2} .$$

        When both $\gamma_1$ and $\gamma_2$ are greater than 0 dB, and if $\gamma_1 = \gamma_2$, then
        $\gamma = \gamma_1 / 2 = \gamma_2 / 2$. If one is much less than the other, then $\gamma$ is approximately
        equal to the smaller one. When both are less than 0 dB, the composite SNR is approximately
        $\gamma = \gamma_1 \gamma_2$.

        .. note::
            The constant terms from Stein's original equations were rearranged. The factor of 2 was removed and
            incorporated into the CRLB equations. This was done so that when $\gamma_2 = \infty$ then
            $\gamma = \gamma_1$.

    References:
        - `Seymour Stein, Algorithms for Ambiguity Function Processing <https://ieeexplore.ieee.org/document/1163621>`_

    Examples:
        Calculate the composite SNR of two signals with equal SNRs. Notice for positive input SNR, the composite SNR
        is linear with slope 1 and intercept -3 dB. For negative input SNR, the composite SNR is linear with slope 2
        and intercept 0 dB.

        .. ipython:: python

            snr1 = np.linspace(-60, 60, 101)

            @savefig sdr_composite_snr_1.png
            plt.figure(); \
            plt.plot(snr1, sdr.composite_snr(snr1, snr1)); \
            plt.xlim(-60, 60); \
            plt.ylim(-60, 60); \
            plt.xlabel("Input SNRs (dB), $\gamma_1$ and $\gamma_2$"); \
            plt.ylabel(r"Composite SNR (dB), $\gamma$"); \
            plt.title("Composite SNR of two signals with equal SNRs");

        Calculate the composite SNR of two signals with different SNRs. Notice the knee of the curve is located at
        `max(0, snr2)`. Left of the knee, the composite SNR is linear with slope 1 and intercept `snr2`.
        Right of the knee, the composite SNR is linear with slope 0 and intercept `snr2`.

        .. ipython:: python

            snr1 = np.linspace(-60, 60, 101)

            plt.figure();
            for snr2 in np.arange(-40, 40 + 10, 10):
                plt.plot(snr1, sdr.composite_snr(snr1, snr2), label=snr2)
            @savefig sdr_composite_snr_2.png
            plt.legend(title="SNR of signal 2 (dB), $\gamma_2$"); \
            plt.xlim(-60, 60); \
            plt.ylim(-60, 60); \
            plt.xlabel("SNR of signal 1 (dB), $\gamma_1$"); \
            plt.ylabel(r"Composite SNR (dB), $\gamma$"); \
            plt.title("Composite SNR of two signals with different SNRs");

    Group:
        estimation-snr
    """
    snr1 = verify_arraylike(snr1, float=True)
    snr2 = verify_arraylike(snr2, float=True)

    # Convert to linear
    snr1 = linear(snr1)
    snr2 = linear(snr2)

    inv_snr = 1 / snr1 + 1 / snr2 + 1 / (snr1 * snr2)
    snr = 1 / inv_snr

    # Convert back to dB
    snr = db(snr)

    return convert_output(snr)



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib.ticker import MaxNLocator
from typing_extensions import Literal

from .._conversion import db
from .._helper import verify_scalar
from ._utility import stem


def real_or_complex_plot(
    x: npt.NDArray,
    y: npt.NDArray,
    ax: plt.Axes,
    diff: Literal["color", "line"] = "color",
    **kwargs,
):
    label = kwargs.pop("label", None)
    if np.iscomplexobj(y):
        if label is None:
            label, label2 = "real", "imag"
        else:
            label, label2 = label + " (real)", label + " (imag)"

        if diff == "color":
            ax.plot(x, y.real, label=label, **kwargs)
            ax.plot(x, y.imag, label=label2, **kwargs)
        elif diff == "line":
            (real,) = ax.plot(x, y.real, "-", label=label, **kwargs)
            kwargs.pop("color", None)
            ax.plot(x, y.imag, "--", color=real.get_color(), label=label2, **kwargs)
        else:
            raise ValueError(f"Argument 'diff' must be 'color' or 'line', not {diff}.")
    else:
        ax.plot(x, y, label=label, **kwargs)

    if label:
        ax.legend()


def standard_plot(
    x: npt.NDArray,
    y: npt.NDArray,
    ax: plt.Axes,
    type: Literal["plot", "stem"] = "plot",
    y_axis: Literal["complex", "mag", "mag^2", "db"] = "mag",
    diff: Literal["color", "line"] = "color",
    **kwargs,
):
    if y_axis == "complex":
        pass
    elif y_axis == "mag":
        y = np.abs(y)
    elif y_axis == "mag^2":
        y = np.abs(y) ** 2
    elif y_axis == "db":
        y = db(np.abs(y) ** 2)
    else:
        raise ValueError(f"Argument 'y_axis' must be 'complex', 'mag', 'mag^2', or 'db', not {y_axis!r}.")

    label = kwargs.pop("label", None)

    if type == "plot":
        if np.iscomplexobj(y):
            if label is None:
                label, label2 = "real", "imag"
            else:
                label, label2 = label + " (real)", label + " (imag)"

            if diff == "color":
                ax.plot(x, y.real, label=label, **kwargs)
                ax.plot(x, y.imag, label=label2, **kwargs)
            elif diff == "line":
                (real,) = ax.plot(x, y.real, "-", label=label, **kwargs)
                kwargs.pop("color", None)
                ax.plot(x, y.imag, "--", color=real.get_color(), label=label2, **kwargs)
            else:
                raise ValueError(f"Argument 'diff' must be 'color' or 'line', not {diff}.")
        else:
            ax.plot(x, y, label=label, **kwargs)

    elif type == "stem":
        stem(x, y, ax=ax, label=label, **kwargs)

    else:
        raise ValueError(f"Argument 'type' must be 'plot' or 'stem', not {type!r}.")

    if label:
        ax.legend()


def verify_sample_rate(sample_rate: float | None, default=1.0):
    if sample_rate is None:
        sample_rate_provided = False
        sample_rate = default
    else:
        sample_rate_provided = True
        verify_scalar(sample_rate, float=True, positive=True)

    return sample_rate, sample_rate_provided


def integer_x_axis(ax: plt.Axes):
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.xaxis.set_minor_locator(MaxNLocator(integer=True))


def min_ylim(y: npt.NDArray, separation: float, sample_rate: float):
    ymin, ymax = plt.gca().get_ylim()
    if ymax - ymin < separation / 4:
        # Find the mean of the signal rounded to the nearest sample
        mean = int(round(np.nanmean(y) / sample_rate))
        ymin = mean - separation / 2
        ymax = mean + separation / 2
        plt.gca().set_ylim(ymin, ymax)



"""
A module containing maximum-length sequences.
"""

from __future__ import annotations

import itertools
from typing import Any, Iterator, overload

import galois
import numpy as np
import numpy.typing as npt
from galois import Poly
from galois.typing import PolyLike
from typing_extensions import Literal

from .._helper import export, verify_condition, verify_literal, verify_scalar
from ._conversion import code_to_sequence, field_to_code
from ._lfsr import FLFSR


@overload
def m_sequence(
    degree: int,
    poly: PolyLike | None = None,
    index: int = 1,
    output: Literal["decimal"] = "decimal",
) -> npt.NDArray[np.int_]: ...


@overload
def m_sequence(
    degree: int,
    poly: PolyLike | None = None,
    index: int = 1,
    output: Literal["field"] = "decimal",
) -> galois.FieldArray: ...


@overload
def m_sequence(
    degree: int,
    poly: PolyLike | None = None,
    index: int = 1,
    output: Literal["bipolar"] = "decimal",
) -> npt.NDArray[np.float64]: ...


@export
def m_sequence(
    degree: Any,
    poly: Any = None,
    index: Any = 1,
    output: Any = "decimal",
) -> Any:
    r"""
    Generates a maximum-length sequence ($m$-sequence) from a Fibonacci linear feedback shift register (LFSR).

    Arguments:
        degree: The degree $m$ of the LFSR.
        poly: The characteristic polynomial of the LFSR over $\mathrm{GF}(q)$. The default is `None`, which uses the
            primitive polynomial of degree $m$ over $\mathrm{GF}(2)$, `galois.primitive_poly(2, degree)`.
        index: The index $i$ in $[1, q^m)$ of the $m$-sequence. The index represents the initial state of the LFSR.
            The index dictates the phase of the $m$-sequence. The integer index is interpreted as a polynomial over
            $\mathrm{GF}(q)$, whose coefficients are the shift register values. The default is 1, which corresponds
            to the $[0, \dots, 0, 1]$ state.
        output: The output format of the $m$-sequence.

            - `"decimal"` (default): The $m$-sequence with decimal values in $[0, q^m)$.
            - `"field"`: The $m$-sequence as a Galois field array over $\mathrm{GF}(q^m)$.
            - `"bipolar"`: The $m$-sequence with bipolar values of 1 and -1. Only valid for $q = 2$.

    Returns:
        The length-$q^m - 1$ $m$-sequence.

    References:
        - https://en.wikipedia.org/wiki/Maximum_length_sequence

    Examples:
        Generate a maximum-length sequence of degree-4 over $\mathrm{GF}(2)$.

        .. ipython:: python

            sdr.m_sequence(4)
            sdr.m_sequence(4, output="bipolar")
            sdr.m_sequence(4, output="field")

        Compare the sequence with index 1 to the sequence with index 2. They are just phase shifts of each other.

        .. ipython:: python

            sdr.m_sequence(4, index=2)

        Generate a maximum-length sequence of degree-4 over $\mathrm{GF}(3^2)$.

        .. ipython:: python

            c = galois.primitive_poly(3**2, 4); c
            x = sdr.m_sequence(4, poly=c); x
            x.size

        Plot the auto-correlation of a length-63 $m$-sequence. Notice that the linear correlation produces sidelobes
        for non-zero lag. However, the circular correlation only produces magnitudes of 1 for non-zero lag.

        .. ipython:: python

            x = sdr.m_sequence(6, output="bipolar")

            @savefig sdr_m_sequence_1.png
            plt.figure(); \
            sdr.plot.correlation(x, x, mode="circular"); \
            plt.ylim(0, 63);

        The cross-correlation of two $m$-sequences with different indices is low for zero lag. However, for non-zero
        lag the cross-correlation is very large.

        .. ipython:: python

            x = sdr.m_sequence(6, index=1, output="bipolar")
            y = sdr.m_sequence(6, index=30, output="bipolar")

            @savefig sdr_m_sequence_2.png
            plt.figure(); \
            sdr.plot.correlation(x, y, mode="circular"); \
            plt.ylim(0, 63);

    Group:
        sequences-maximum-length
    """
    verify_scalar(degree, int=True, positive=True)

    if poly is None:
        poly = galois.primitive_poly(2, degree)  # Characteristic polynomial
    elif isinstance(poly, Poly):
        pass
    else:
        poly = Poly._PolyLike(poly, field=galois.GF(2))
    if not poly.degree == degree:
        raise ValueError(f"Argument 'poly' must be a polynomial of degree {degree}, not {poly.degree}.")
    if not poly.is_primitive():
        raise ValueError(f"Argument 'poly' must be a primitive polynomial, {poly} is not.")
    q = poly.field.order

    verify_scalar(index, int=True, inclusive_min=1, exclusive_max=q**degree)
    verify_literal(output, ["decimal", "field", "bipolar"])

    state_poly = Poly.Int(index, field=poly.field)
    state_vector = state_poly.coefficients(degree)
    lfsr = FLFSR(poly, state=state_vector)
    code = lfsr.step(q**degree - 1)

    if output == "decimal":
        return field_to_code(code)
    elif output == "field":
        return code
    else:
        if not q == 2:
            raise ValueError(f"Argument 'output' can only be 'bipolar' when q = 2, not {q}.")
        return code_to_sequence(field_to_code(code))


@export
def preferred_pairs(
    degree: int,
    poly: PolyLike | None = None,
) -> Iterator[tuple[Poly, Poly]]:
    r"""
    Generates primitive polynomials of degree $m$ that produce preferred pair $m$-sequences.

    Arguments:
        degree: The degree $m$ of the $m$-sequences.
        poly: The first polynomial $f(x)$ in the preferred pair. If `None`, all primitive polynomials of degree $m$
            that yield preferred pair $m$-sequences are returned.

    Returns:
        An iterator of primitive polynomials $(f(x), g(x))$ of degree $m$ that produce preferred pair $m$-sequences.

    See Also:
        sdr.is_preferred_pair, sdr.gold_code

    Notes:
        A preferred pair of primitive polynomials of degree $m$ are two polynomials $f(x)$ and $g(x)$ such that the
        periodic cross-correlation of the two $m$-sequences generated by $f(x)$ and $g(x)$ have only the values in
        $\{-1, -t(m), t(m) - 2\}$.

        $$
        t(m) = \begin{cases}
        2^{(m+1)/2} + 1 & \text{if $m$ is odd} \\
        2^{(m+2)/2} + 1 & \text{if $m$ is even}
        \end{cases}
        $$

        There are no preferred pairs for degrees divisible by 4.

    References:
        - John Proakis, *Digital Communications*, Chapter 12.2-5: Generation of PN Sequences.

    Examples:
        Generate one preferred pair with $f(x) = x^5 + x^3 + 1$.

        .. ipython:: python

            next(sdr.preferred_pairs(5, poly="x^5 + x^3 + 1"))

        Generate all preferred pairs with $f(x) = x^5 + x^3 + 1$.

        .. ipython:: python

            list(sdr.preferred_pairs(5, poly="x^5 + x^3 + 1"))

        Generate all preferred pairs with degree 5.

        .. ipython:: python

            list(sdr.preferred_pairs(5))

        Note, there are no preferred pairs for degrees divisible by 4.

        .. ipython:: python

            list(sdr.preferred_pairs(4))
            list(sdr.preferred_pairs(8))

    Group:
        sequences-maximum-length
    """
    verify_scalar(degree, int=True, positive=True)

    if degree % 4 == 0:
        # There are no preferred pairs for degrees divisible by 4
        return

    # Compute t(m) for degree m, Equation 12.2-73
    if degree % 2 == 1:
        t_m = 2 ** ((degree + 1) // 2) + 1
    else:
        t_m = 2 ** ((degree + 2) // 2) + 1

    # Determine the valid cross-correlation values for preferred pairs, Page 799
    valid_values = [-1, -t_m, t_m - 2]

    if poly is None:
        # Find all combinations of primitive polynomials of degree m
        for poly1, poly2 in itertools.combinations(galois.primitive_polys(2, degree), 2):
            # Create first m-sequence with the first polynomial
            u = m_sequence(degree, poly1, output="bipolar")
            U = np.fft.fft(u)

            # Create second m-sequence with the second polynomial
            v = m_sequence(degree, poly2, output="bipolar")
            V = np.fft.fft(v)

            # Compute the periodic cross-correlation of the two m-sequences
            pccf = np.fft.ifft(U * V.conj()).real  # The inputs are real, so the output is real
            pccf = np.around(pccf).astype(int)  # Round to the nearest integer so isin() works

            if np.all(np.isin(pccf, valid_values)):
                yield poly1, poly2
    else:
        # Find all combinations of the first polynomial with all primitive polynomials of degree m
        poly1 = Poly._PolyLike(poly, field=galois.GF(2))
        if not poly1.degree == degree:
            raise ValueError(f"Argument 'poly1' must be a polynomial of degree {degree}, not {poly1.degree}.")
        if not poly1.is_primitive():
            raise ValueError(f"Argument 'poly1' must be a primitive polynomial, {poly1} is not.")

        # Create first m-sequence with the first polynomial
        u = m_sequence(degree, poly1, output="bipolar")
        U = np.fft.fft(u)

        for poly2 in galois.primitive_polys(2, degree):
            # Create second m-sequence with the second polynomial
            v = m_sequence(degree, poly2, output="bipolar")
            V = np.fft.fft(v)

            # Compute the periodic cross-correlation of the two m-sequences
            pccf = np.fft.ifft(U * V.conj()).real  # The inputs are real, so the output is real
            pccf = np.around(pccf).astype(int)  # Round to the nearest integer so isin() works

            if np.all(np.isin(pccf, valid_values)):
                yield poly1, poly2


@export
def is_preferred_pair(
    poly1: PolyLike,
    poly2: PolyLike,
) -> bool:
    r"""
    Determines if two primitive polynomials generate preferred pair $m$-sequences.

    Arguments:
        poly1: The first primitive polynomial $f(x)$.
        poly2: The second primitive polynomial $g(x)$.

    Returns:
        A boolean indicating if the two primitive polynomials generate preferred pair $m$-sequences.

    See Also:
        sdr.preferred_pairs, sdr.gold_code

    Notes:
        A preferred pair of primitive polynomials of degree $m$ are two polynomials $f(x)$ and $g(x)$ such that the
        periodic cross-correlation of the two $m$-sequences generated by $f(x)$ and $g(x)$ have only the values in
        $\{-1, -t(m), t(m) - 2\}$.

        $$
        t(m) = \begin{cases}
        2^{(m+1)/2} + 1 & \text{if $m$ is odd} \\
        2^{(m+2)/2} + 1 & \text{if $m$ is even}
        \end{cases}
        $$

        There are no preferred pairs for degrees divisible by 4.

    References:
        - John Proakis, *Digital Communications*, Chapter 12.2-5: Generation of PN Sequences.

    Examples:
        Determine if the pair $f(x) = x^5 + x^3 + 1$ and $g(x) = x^5 + x^3 + x^2 + x + 1$ is a preferred pair.

        .. ipython:: python

            sdr.is_preferred_pair("x^5 + x^3 + 1", "x^5 + x^3 + x^2 + x + 1")

        Determine if the pair $f(x) = x^5 + x^3 + 1$ and $g(x) = x^5 + x^2 + 1$ is a preferred pair.

        .. ipython:: python

            sdr.is_preferred_pair("x^5 + x^3 + 1", "x^5 + x^2 + 1")

    Group:
        sequences-maximum-length
    """
    poly1 = Poly._PolyLike(poly1, field=galois.GF(2))
    poly2 = Poly._PolyLike(poly2, field=galois.GF(2))

    verify_condition(poly1.degree == poly2.degree)
    verify_condition(poly1.is_primitive())
    verify_condition(poly2.is_primitive())

    # if not poly1.degree == poly2.degree:
    #     raise ValueError(
    #         f"Arguments 'poly1' and 'poly2' must have the same degree, not {poly1.degree} and {poly2.degree}."
    #     )
    # if not poly1.is_primitive():
    #     raise ValueError(f"Argument 'poly1' must be a primitive polynomial, {poly1} is not.")
    # if not poly2.is_primitive():
    #     raise ValueError(f"Argument 'poly2' must be a primitive polynomial, {poly2} is not.")

    degree = poly1.degree
    if degree % 4 == 0:
        # There are no preferred pairs for degrees divisible by 4
        return False

    # Compute t(m) for degree m, Equation 12.2-73
    if degree % 2 == 1:
        t_m = 2 ** ((degree + 1) // 2) + 1
    else:
        t_m = 2 ** ((degree + 2) // 2) + 1

    # Determine the valid cross-correlation values for preferred pairs, Page 799
    valid_values = [-1, -t_m, t_m - 2]

    # Create first m-sequence with the first polynomial
    u = m_sequence(degree, poly1, output="bipolar")
    U = np.fft.fft(u)

    # Create second m-sequence with the second polynomial
    v = m_sequence(degree, poly2, output="bipolar")
    V = np.fft.fft(v)

    # Compute the periodic cross-correlation of the two m-sequences
    pccf = np.fft.ifft(U * V.conj()).real  # The inputs are real, so the output is real
    pccf = np.around(pccf).astype(int)  # Round to the nearest integer so isin() works

    valid = bool(np.all(np.isin(pccf, valid_values)))

    return valid



"""
A module containing various voltage measurement functions.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._conversion import db as to_db
from .._helper import convert_output, export, verify_arraylike, verify_bool


@export
def peak_voltage(x: npt.ArrayLike, db: bool = False) -> float:
    r"""
    Measures the peak voltage of a time-domain signal $x[n]$.

    $$V_{\text{peak}} = \max \left| x[n] \right|$$

    Arguments:
        x: The time-domain signal $x[n]$ to measure.
        db: Indicates whether to return the result in decibels (dB).

    Returns:
        The peak voltage. If `db=False`, $V_{\text{peak}}$ is returned.
        If `db=True`, $20 \log_{10} V_{\text{peak}}$ is returned.

    Group:
        measurement-voltage
    """
    x = verify_arraylike(x, complex=True)
    verify_bool(db)

    V_peak = np.max(np.abs(x))
    if db:
        V_peak = to_db(V_peak, type="voltage")

    return convert_output(V_peak)


@export
def rms_voltage(x: npt.ArrayLike, db: bool = False) -> float:
    r"""
    Measures the root-mean-square (RMS) voltage of a time-domain signal $x[n]$.

    $$V_{\text{rms}} = \sqrt{\frac{1}{N} \sum_{n=0}^{N-1} \left| x[n] \right|^2}$$

    Arguments:
        x: The time-domain signal $x[n]$ to measure.
        db: Indicates whether to return the result in decibels (dB).

    Returns:
        The root-mean-square voltage. If `db=False`, $V_{\text{rms}}$ is returned.
        If `db=True`, $20 \log_{10} V_{\text{rms}}$ is returned.

    Group:
        measurement-voltage
    """
    x = verify_arraylike(x, complex=True)
    verify_bool(db)

    V_rms = np.sqrt(np.mean(np.abs(x) ** 2))
    if db:
        V_rms = to_db(V_rms, type="voltage")

    return convert_output(V_rms)


@export
def crest_factor(x: npt.ArrayLike) -> float:
    r"""
    Measures the crest factor of a time-domain signal $x[n]$.

    $$\text{CF} = \frac{V_{\text{peak}}}{V_{\text{rms}}}$$

    Arguments:
        x: The time-domain signal $x[n]$ to measure.

    Returns:
        The crest factor of $x[n]$.

    See Also:
        sdr.peak_voltage, sdr.rms_voltage

    References:
        - https://en.wikipedia.org/wiki/Crest_factor

    Group:
        measurement-voltage
    """
    return peak_voltage(x) / rms_voltage(x)



"""
A module for specific filter applications.
"""

from __future__ import annotations

import warnings

import numpy as np
import scipy.signal
from typing_extensions import Literal

from .._helper import export, verify_bool, verify_literal, verify_scalar
from ._fir import FIR
from ._iir import IIR


@export
class MovingAverager(FIR):
    r"""
    Implements a moving average FIR filter.

    Notes:
        A discrete-time moving average with length $L$ is an FIR filter with impulse response

        $$h[n] = \frac{1}{L}, \quad 0 \le n \le L - 1 .$$

    Examples:
        Create an FIR moving average filter and an IIR leaky integrator filter.

        .. ipython:: python

            fir = sdr.MovingAverager(30)
            iir = sdr.LeakyIntegrator(1 - 2 / 30)

        Compare the step responses.

        .. ipython:: python

            @savefig sdr_MovingAverager_1.png
            plt.figure(); \
            sdr.plot.step_response(fir, N=100, label="Moving Averager"); \
            sdr.plot.step_response(iir, N=100, label="Leaky Integrator");

        Compare the magnitude responses.

        .. ipython:: python

            @savefig sdr_MovingAverager_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(fir, label="Moving Averager"); \
            sdr.plot.magnitude_response(iir, label="Leaky Integrator"); \
            plt.ylim(-35, 5);

        Compare the output of the two filters to a Gaussian random process.

        .. ipython:: python

            x = np.random.randn(1_000) + 2.0; \
            y_fir = fir(x); \
            y_iir = iir(x)

            @savefig sdr_MovingAverager_3.png
            plt.figure(); \
            sdr.plot.time_domain(y_fir, label="Moving Averager"); \
            sdr.plot.time_domain(y_iir, label="Leaky Integrator");

    Group:
        dsp-filter-applications
    """

    def __init__(
        self,
        length: int,
        streaming: bool = False,
    ):
        """
        Creates a moving average FIR filter.

        Arguments:
            length: The length of the moving average filter $L$.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs are
                preserved between calls to :meth:`~sdr.MovingAverager.__call__()`.

        Examples:
            See the :ref:`fir-filters` example.
        """
        verify_scalar(length, int=True, positive=True)
        verify_bool(streaming)

        h = np.ones(length) / length

        super().__init__(h, streaming=streaming)


@export
class Differentiator(FIR):
    r"""
    Implements a differentiator FIR filter.

    Notes:
        A discrete-time differentiator is an FIR filter with impulse response

        $$h[n] = \frac{(-1)^n}{n} \cdot h_{win}[n], \quad -\frac{N}{2} \le n \le \frac{N}{2} .$$

        The truncated impulse response is multiplied by the windowing function $h_{win}[n]$.

    References:
        - Michael Rice, *Digital Communications: A Discrete Time Approach*, Section 3.3.3.

    Examples:
        Create a differentiator FIR filter.

        .. ipython:: python

            fir = sdr.Differentiator()

        Differentiate a Gaussian pulse.

        .. ipython:: python

            x = sdr.gaussian(0.3, 5, 10); \
            y = fir(x, "same")

            @savefig sdr_Differentiator_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, label="Input"); \
            sdr.plot.time_domain(y, label="Derivative"); \
            plt.title("Discrete-time differentiation of a Gaussian pulse");

        Differentiate a raised cosine pulse.

        .. ipython:: python

            x = sdr.root_raised_cosine(0.1, 8, 10); \
            y = fir(x, "same")

            @savefig sdr_Differentiator_2.png
            plt.figure(); \
            sdr.plot.time_domain(x, label="Input"); \
            sdr.plot.time_domain(y, label="Derivative"); \
            plt.title("Discrete-time differentiation of a raised cosine pulse");

        Plot the frequency response across filter order.

        .. ipython:: python

            fir_2 = sdr.Differentiator(2); \
            fir_6 = sdr.Differentiator(6); \
            fir_10 = sdr.Differentiator(10); \
            fir_20 = sdr.Differentiator(20); \
            fir_40 = sdr.Differentiator(40); \
            fir_80 = sdr.Differentiator(80)

            @savefig sdr_Differentiator_3.png
            plt.figure(); \
            sdr.plot.magnitude_response(fir_2, y_axis="linear", label="N=2"); \
            sdr.plot.magnitude_response(fir_6, y_axis="linear", label="N=6"); \
            sdr.plot.magnitude_response(fir_10, y_axis="linear", label="N=10"); \
            sdr.plot.magnitude_response(fir_20, y_axis="linear", label="N=20"); \
            sdr.plot.magnitude_response(fir_40, y_axis="linear", label="N=40"); \
            sdr.plot.magnitude_response(fir_80, y_axis="linear", label="N=80"); \
            f = np.linspace(0, 0.5, 100); \
            plt.plot(f, np.abs(2 * np.pi * f)**2, color="k", linestyle="--", label="Theory"); \
            plt.legend(); \
            plt.title("Magnitude response of differentiator FIR filters");

    Group:
        dsp-filter-applications
    """

    def __init__(
        self,
        order: int = 20,
        window: str | float | tuple | None = "blackman",
        streaming: bool = False,
    ):
        """
        Creates a differentiator FIR filter.

        Arguments:
            order: The order of the FIR differentiator $N$. The filter length is $N + 1$.
                Increasing the filter order increases the bandwidth of the differentiator.
            window: The SciPy window definition. See :func:`scipy.signal.windows.get_window` for details.
                If `None`, no window is applied.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs are
                preserved between calls to :meth:`~sdr.Differentiator.__call__`.

        Examples:
            See the :ref:`fir-filters` example.
        """
        verify_scalar(order, int=True, positive=True, even=True)
        verify_bool(streaming)

        n = np.arange(-order // 2, order // 2 + 1)  # Sample index centered about 0
        with np.errstate(divide="ignore"):
            h = (-1.0) ** n / n  # Impulse response
        h[order // 2] = 0

        if window is not None:
            h *= scipy.signal.windows.get_window(window, order + 1, fftbins=False)

        super().__init__(h, streaming=streaming)

    # TODO: Use np.diff() if it is faster


@export
class Integrator(IIR):
    r"""
    Implements an integrator IIR filter.

    Notes:
        A discrete-time integrator is an IIR filter that continuously accumulates the input signal.
        Accordingly, it has infinite gain at DC.

        The backward integrator is defined by:

        $$y[n] = y[n-1] + x[n-1]$$
        $$H(z) = \frac{z^{-1}}{1 - z^{-1}}$$

        The trapezoidal integrator is defined by:

        $$y[n] = y[n-1] + \frac{1}{2}x[n] + \frac{1}{2}x[n-1]$$
        $$H(z) = \frac{1}{2} \frac{1 + z^{-1}}{1 - z^{-1}}$$

        The forward integrator is defined by:

        $$y[n] = y[n-1] + x[n]$$
        $$H(z) = \frac{1}{1 - z^{-1}}$$

    Examples:
        Create integrating IIR filters.

        .. ipython:: python

            iir_back = sdr.Integrator("backward"); \
            iir_trap = sdr.Integrator("trapezoidal"); \
            iir_forw = sdr.Integrator("forward")

        Integrate a Gaussian pulse.

        .. ipython:: python

            x = sdr.gaussian(0.3, 5, 10); \
            y_back = iir_back(x); \
            y_trap = iir_trap(x); \
            y_forw = iir_forw(x)

            @savefig sdr_Integrator_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, label="Input"); \
            sdr.plot.time_domain(y_back, label="Integral (backward)"); \
            sdr.plot.time_domain(y_trap, label="Integral (trapezoidal)"); \
            sdr.plot.time_domain(y_forw, label="Integral (forward)"); \
            plt.title("Discrete-time integration of a Gaussian pulse");

        Integrate a raised cosine pulse.

        .. ipython:: python

            x = sdr.root_raised_cosine(0.1, 8, 10); \
            y_back = iir_back(x); \
            y_trap = iir_trap(x); \
            y_forw = iir_forw(x)

            @savefig sdr_Integrator_2.png
            plt.figure(); \
            sdr.plot.time_domain(x, label="Input"); \
            sdr.plot.time_domain(y_back, label="Integral (backward)"); \
            sdr.plot.time_domain(y_trap, label="Integral (trapezoidal)"); \
            sdr.plot.time_domain(y_forw, label="Integral (forward)"); \
            plt.title("Discrete-time integration of a raised cosine pulse");

        Plot the frequency responses and compare them to theory.

        .. ipython:: python
            :okwarning:

            f = np.linspace(0, 0.5, 100)
            H_theory = 1/(2 * np.pi * f)

        .. ipython:: python

            @savefig sdr_Integrator_3.png
            plt.figure(); \
            sdr.plot.magnitude_response(iir_back, label="Backward"); \
            sdr.plot.magnitude_response(iir_trap, label="Trapezoidal"); \
            sdr.plot.magnitude_response(iir_forw, label="Forward"); \
            plt.plot(f, sdr.db(np.abs(H_theory)**2), color="k", linestyle="--", label="Theory"); \
            plt.legend(); \
            plt.title("Magnitude response of integrating IIR filters");

    Group:
        dsp-filter-applications
    """

    def __init__(
        self,
        method: Literal["backward", "trapezoidal", "forward"] = "trapezoidal",
        streaming: bool = False,
    ):
        """
        Creates an integrating IIR filter.

        Arguments:
            method: The integration method.

                - `"backward"`: Rectangular integration with height $x[n-1]$.
                - `"trapezoidal"`: Trapezoidal integration with heights $x[n-1]$ and $x[n]$.
                - `"forward"`: Rectangular integration with height $x[n]$.

            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs and outputs are
                preserved between calls to :meth:`~Integrator.__call__()`.

        Examples:
            See the :ref:`iir-filters` example.
        """
        verify_literal(method, ["backward", "trapezoidal", "forward"])
        verify_bool(streaming)

        if method == "backward":
            with warnings.catch_warnings():
                # SciPy throws a warning because there is a leading 0
                warnings.simplefilter("ignore", scipy.signal.BadCoefficients)
                super().__init__([0, 1], [1, -1], streaming=streaming)
        elif method == "forward":
            super().__init__([1], [1, -1], streaming=streaming)
        elif method == "trapezoidal":
            super().__init__([0.5, 0.5], [1, -1], streaming=streaming)

    # TODO: Use np.cumsum() if it is faster


@export
class LeakyIntegrator(IIR):
    r"""
    Implements a leaky integrator IIR filter.

    Notes:
        A discrete-time leaky integrator is an IIR filter that approximates an FIR moving average.
        The previous output is remembered with the leaky factor $\alpha$ and the new input is scaled with $1 - \alpha$.

        The difference equation is

        $$y[n] = \alpha \cdot y[n-1] + (1 - \alpha) \cdot x[n] .$$

        The transfer functions is

        $$H(z) = \frac{1 - \alpha}{1 - \alpha z^{-1}} .$$

        .. code-block:: text
            :caption: IIR Integrator Block Diagram

                  1 - α
            x[n] ------->@---------------+--> y[n]
                         ^               |
                       α |   +------+    |
                         +---| z^-1 |<---+
                             +------+

    Examples:
        Create an FIR moving average filter and an IIR leaky integrator filter.

        .. ipython:: python

            fir = sdr.MovingAverager(30)
            iir = sdr.LeakyIntegrator(1 - 2 / 30)

        Compare the step responses.

        .. ipython:: python

            @savefig sdr_LeakyIntegrator_1.png
            plt.figure(); \
            sdr.plot.step_response(fir, N=100, label="Moving Averager"); \
            sdr.plot.step_response(iir, N=100, label="Leaky Integrator");

        Compare the magnitude responses.

        .. ipython:: python

            @savefig sdr_LeakyIntegrator_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(fir, label="Moving Averager"); \
            sdr.plot.magnitude_response(iir, label="Leaky Integrator"); \
            plt.ylim(-35, 5);

        Compare the output of the two filters to a Gaussian random process.

        .. ipython:: python

            x = np.random.randn(1_000) + 2.0; \
            y_fir = fir(x); \
            y_iir = iir(x)

            @savefig sdr_LeakyIntegrator_3.png
            plt.figure(); \
            sdr.plot.time_domain(y_fir, label="Moving Averager"); \
            sdr.plot.time_domain(y_iir, label="Leaky Integrator");

    Group:
        dsp-filter-applications
    """

    def __init__(
        self,
        alpha: float,
        streaming: bool = False,
    ):
        r"""
        Creates a leaky integrator IIR filter.

        Arguments:
            alpha: The leaky factor $\alpha$. An FIR moving average with length $L$ is approximated when
                $\alpha = 1 - 2/L$.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs and outputs are
                preserved between calls to :meth:`~LeakyIntegrator.__call__()`.

        Examples:
            See the :ref:`iir-filters` example.
        """
        verify_scalar(alpha, float=True, exclusive_min=0, exclusive_max=1)
        verify_bool(streaming)

        b = [1 - alpha]
        a = [1, -alpha]

        super().__init__(b, a, streaming=streaming)

    # TODO: Use np.cumsum() if it is faster



import numpy as np

import sdr


def test_1():
    """
    MATLAB:
        >> x = 0:39;
        >> downsample(x, 4)'
    """
    x = np.arange(40)
    y = sdr.downsample(x, 4)
    y_truth = np.array([0, 4, 8, 12, 16, 20, 24, 28, 32, 36])
    assert np.array_equal(y, y_truth)



"""
A module containing a class that implements an energy detector.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.stats

from .._conversion import linear
from .._helper import convert_output, export, verify_arraylike, verify_scalar


@export
class EnergyDetector:
    r"""
    Implements an energy detector.

    Notes:
        The null and alternative hypotheses are given by the following. The signal $s[n]$ is assumed to be
        a complex-valued random process with noise variance $\sigma_s^2$. The noise $w[n]$ is assumed to be
        a complex-valued random process with noise variance $\sigma^2$.

        $$\mathcal{H}_0: x[n] = w[n]$$
        $$\mathcal{H}_1: x[n] = s[n] + w[n]$$

        The test statistic $T(x)$ is given by:

        $$T(x) = \sum\limits_{n=0}^{N-1} \left| x[n] \right|^2 > \gamma'$$
        $$
        \frac{T(x)}{\sigma^2 / 2} \sim \chi_{2N}^2 & \text{ under } \mathcal{H}_0 \\
        \frac{T(x)}{(\sigma_s^2 + \sigma^2) / 2} \sim \chi_{2N}^2 & \text{ under } \mathcal{H}_1
        $$

        The probability of detection $P_d$, probability of false alarm $P_{fa}$, and detection threshold
        $\gamma'$ are given by:

        $$P_d = Q_{\chi_{2N}^2}\left( \frac{Q_{\chi_{2N}^2}^{-1}(P_{fa})}{\sigma_s^2 /\sigma^2 + 1} \right)$$
        $$P_{fa} = Q_{\chi_{2N}^2}\left( \frac{\gamma'}{\sigma^2 / 2} \right)$$
        $$\gamma' = \frac{\sigma^2}{2} Q_{\chi_N^2}^{-1}(P_{fa})$$

    References:
        - Steven Kay, *Fundamentals of Statistical Signal Processing: Detection Theory*, Sections 5.3.

    Group:
        detection-detectors
    """

    # def __init__(
    #     self,
    #     N_nc: int,
    #     p_fa: float = 1e-6,
    #     streaming: bool = False,
    # ):
    #     r"""
    #     Initializes the energy detector.

    #     Arguments:
    #         N_nc: The number of samples $N_{nc}$ to non-coherently integrate.
    #         p_fa: The desired probability of false alarm $P_{fa}$.
    #     """
    #     if not isinstance(N_nc, int):
    #         raise TypeError(f"Argument 'N_nc' must be an integer, not {type(N_nc)}.")
    #     if not N_nc >= 1:
    #         raise ValueError(f"Argument 'N_nc' must be greater than or equal to 1, not {N_nc}.")
    #     self._N_nc = N_nc

    #     if not isinstance(p_fa, float):
    #         raise TypeError(f"Argument 'p_fa' must be a float, not {type(p_fa)}.")
    #     if not 0 < p_fa < 1:
    #         raise ValueError(f"Argument 'p_fa' must be in the range (0, 1), not {p_fa}.")
    #     self._p_fa = p_fa

    #     if not isinstance(streaming, bool):
    #         raise TypeError(f"Argument 'streaming' must be a bool, not {type(streaming)}.")
    #     self._streaming = streaming

    #     self._fir = FIR(np.ones(N_nc), streaming=streaming)

    @staticmethod
    def roc(
        snr: float,
        N_nc: float,
        p_fa: npt.ArrayLike | None = None,
        complex: bool = True,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        r"""
        Computes the receiver operating characteristic (ROC) curve.

        Arguments:
            snr: The received signal-to-noise ratio $\sigma_s^2 / \sigma^2$ in dB.
            N_nc: The number of samples $N_{nc}$ to non-coherently integrate.
            p_fa: The probability of false alarm $P_{fa}$. If `None`, the ROC curve is computed for
                `p_fa = np.logspace(-10, 0, 101)`.
            complex: Indicates whether the signal is complex.

        Returns:
            - The probability of false alarm $P_{fa}$.
            - The probability of detection $P_d$.

        Examples:
            Plot the theoretical ROC curves for integrating a single sample at various SNRs.

            .. ipython:: python

                @savefig sdr_EnergyDetector_roc_1.png
                plt.figure(); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(-20, 1), label=f"SNR = -20 dB"); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(-10, 1), label=f"SNR = -10 dB"); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(0, 1), label=f"SNR = -0 dB"); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(10, 1), label=f"SNR = 10 dB"); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(20, 1), label=f"SNR = 20 dB");

            Plot the theoretical ROC curves for various integration lengths at -10 dB SNR.

            .. ipython:: python

                @savefig sdr_EnergyDetector_roc_2.png
                plt.figure(); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(-10, 1), label=f"N = 1"); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(-10, 10), label=f"N = 10"); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(-10, 100), label=f"N = 100"); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(-10, 1_000), label=f"N = 1,000"); \
                sdr.plot.roc(*sdr.EnergyDetector.roc(-10, 5_000), label=f"N = 5,000");
        """
        verify_scalar(snr, float=True)
        verify_scalar(N_nc, int=True, positive=True)
        if p_fa is None:
            p_fa = np.logspace(-10, 0, 101)
        else:
            p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)

        p_d = EnergyDetector.p_d(snr, N_nc, p_fa, complex=complex)

        return convert_output(p_fa), convert_output(p_d)

    @staticmethod
    def p_d(
        snr: npt.ArrayLike,
        N_nc: npt.ArrayLike,
        p_fa: npt.ArrayLike,
        complex: bool = True,
    ) -> npt.NDArray[np.float64]:
        r"""
        Computes the probability of detection $P_d$.

        Arguments:
            snr: The received signal-to-noise ratio $\sigma_s^2 / \sigma^2$ in dB.
            N_nc: The number of samples $N_{nc}$ to non-coherently integrate.
            p_fa: The probability of false alarm $P_{fa}$.
            complex: Indicates whether the signal is real or complex.

        Returns:
            The probability of detection $P_d$.

        Notes:
            For real signals:

            $$
            P_d &= Q_{\chi_N^2}\left( \frac{\sigma^2 Q_{\chi_N^2}^{-1}(P_{fa})}{\sigma_s^2 + \sigma^2} \right) \\
            &= Q_{\chi_N^2}\left( \frac{Q_{\chi_N^2}^{-1}(P_{fa})}{\sigma_s^2 /\sigma^2 + 1} \right)
            $$

            For complex signals:

            $$P_d = Q_{\chi_{2N}^2}\left( \frac{Q_{\chi_{2N}^2}^{-1}(P_{fa})}{\sigma_s^2 /\sigma^2 + 1} \right)$$

        References:
            - Steven Kay, *Fundamentals of Statistical Signal Processing: Detection Theory*,
              Equation 5.3.

        Examples:
            .. ipython:: python

                snr = np.linspace(-20, 10, 101)

                @savefig sdr_EnergyDetector_p_d_1.png
                plt.figure(); \
                sdr.plot.p_d(snr, sdr.EnergyDetector.p_d(snr, 25, 1e-1), label="$P_{fa} = 10^{-1}$"); \
                sdr.plot.p_d(snr, sdr.EnergyDetector.p_d(snr, 25, 1e-2), label="$P_{fa} = 10^{-2}$"); \
                sdr.plot.p_d(snr, sdr.EnergyDetector.p_d(snr, 25, 1e-3), label="$P_{fa} = 10^{-3}$"); \
                sdr.plot.p_d(snr, sdr.EnergyDetector.p_d(snr, 25, 1e-4), label="$P_{fa} = 10^{-4}$"); \
                sdr.plot.p_d(snr, sdr.EnergyDetector.p_d(snr, 25, 1e-5), label="$P_{fa} = 10^{-5}$");
        """
        snr = verify_arraylike(snr, float=True)
        N_nc = verify_arraylike(N_nc, int=True, positive=True)
        p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)

        snr_linear = linear(snr)
        if not complex:
            nu = N_nc  # Degrees of freedom
            gamma_dprime = scipy.stats.chi2.isf(p_fa, nu)  # Normalized threshold
            p_d = scipy.stats.chi2.sf(gamma_dprime / (snr_linear + 1), nu)
        else:
            nu = 2 * N_nc  # Degrees of freedom
            gamma_dprime = scipy.stats.chi2.isf(p_fa, nu)  # Normalized threshold
            p_d = scipy.stats.chi2.sf(gamma_dprime / (snr_linear + 1), nu)

        return convert_output(p_d)

    @staticmethod
    def p_fa(
        threshold: npt.ArrayLike,
        N_nc: npt.ArrayLike,
        sigma2: npt.ArrayLike,
        complex: bool = True,
    ) -> npt.NDArray[np.float64]:
        r"""
        Computes the probability of false alarm $P_{fa}$.

        Arguments:
            threshold: The threshold $\gamma'$.
            N_nc: The number of samples $N_{nc}$ to non-coherently integrate.
            sigma2: The noise variance $\sigma^2$.
            complex: Indicates whether the signal is complex.

        Returns:
            The probability of false alarm $P_{fa}$.

        Notes:
            For real signals:

            $$P_{fa} = Q_{\chi_N^2}\left( \frac{\gamma'}{\sigma^2} \right)$$

            For complex signals:

            $$P_{fa} = Q_{\chi_{2N}^2}\left( \frac{\gamma'}{\sigma^2 / 2} \right)$$

        References:
            - Steven Kay, *Fundamentals of Statistical Signal Processing: Detection Theory*,
              Equation 5.2.
        """
        threshold = verify_arraylike(threshold, float=True)
        N_nc = verify_arraylike(N_nc, int=True, positive=True)
        sigma2 = verify_arraylike(sigma2, float=True, non_negative=True)

        if not complex:
            nu = N_nc  # Degrees of freedom
            p_fa = scipy.stats.chi2.sf(threshold / sigma2, nu)
        else:
            nu = 2 * N_nc  # Degrees of freedom
            p_fa = scipy.stats.chi2.sf(threshold / (sigma2 / 2), nu)

        return convert_output(p_fa)

    @staticmethod
    def threshold(
        N_nc: npt.ArrayLike,
        p_fa: npt.ArrayLike,
        sigma2: npt.ArrayLike,
        complex: bool = True,
    ) -> npt.NDArray[np.float64]:
        r"""
        Computes the threshold $\gamma'$.

        Arguments:
            N_nc: The number of samples $N_{nc}$ to non-coherently integrate.
            p_fa: The probability of false alarm $P_{fa}$.
            sigma2: The noise variance $\sigma^2$.
            complex: Indicates whether the signal is complex.

        Returns:
            The threshold $\gamma'$.

        Notes:
            For real signals:

            $$\gamma' = \sigma^2 Q_{\chi_N^2}^{-1}(P_{fa})$$

            For complex signals:

            $$\gamma' = \frac{\sigma^2}{2} Q_{\chi_N^2}^{-1}(P_{fa})$$

        References:
            - Steven Kay, *Fundamentals of Statistical Signal Processing: Detection Theory*,
              Equation 5.2.
        """
        N_nc = verify_arraylike(N_nc, int=True, positive=True)
        p_fa = verify_arraylike(p_fa, float=True, inclusive_min=0, inclusive_max=1)
        sigma2 = verify_arraylike(sigma2, float=True, non_negative=True)

        if not complex:
            nu = N_nc  # Degrees of freedom
            gamma_prime = sigma2 * scipy.stats.chi2.isf(p_fa, nu)
        else:
            nu = 2 * N_nc  # Degrees of freedom
            gamma_prime = sigma2 / 2 * scipy.stats.chi2.isf(p_fa, nu)

        return convert_output(gamma_prime)

    # def test_statistic(self, x: npt.ArrayLike) -> npt.NDArray[np.float64]:
    #     """
    #     Computes the test statistic $T(x)$.

    #     Arguments:
    #         x: The received signal $x[n]$.

    #     Returns:
    #         The test statistic $T(x)$.
    #     """
    #     x = np.asarray(x)

    #     if self.streaming:
    #         return self._fir(np.abs(x) ** 2)

    #     return self._fir(np.abs(x) ** 2, mode="same")

    # def detect(self, x: npt.ArrayLike, sigma2: float) -> npt.NDArray[np.float64]:
    #     r"""
    #     Detects the presence of a signal.

    #     Arguments:
    #         x: The received signal $x[n]$.
    #         sigma2: The noise variance $\sigma^2$.

    #     Returns:
    #         The decision statistic $d$.
    #     """
    #     T = self.test_statistic(x)
    #     gamma = self.threshold(self.N_nc, self.desired_p_fa, sigma2)
    #     return T >= gamma

    # @property
    # def N_nc(self) -> int:
    #     """
    #     The number of samples $N_{nc}$ to non-coherently integrate.
    #     """
    #     return self._N_nc

    # @property
    # def desired_p_fa(self) -> float:
    #     """
    #     The desired probability of false alarm $P_{fa}$.
    #     """
    #     return self._p_fa

    # @property
    # def streaming(self) -> bool:
    #     """
    #     Indicates whether the detector is in streaming mode.
    #     """
    #     return self._streaming



"""
A module to determine axis units.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt


def time_units(time: npt.NDArray[np.float64]) -> tuple[str, float]:
    """
    Determines the appropriate time units to use for a given time array.
    """
    max_time = np.nanmax(np.abs(time))

    if max_time > 1:
        scalar = 1
        units = "s"
    elif max_time > 1e-3:
        scalar = 1e3
        units = "ms"
    elif max_time > 1e-6:
        scalar = 1e6
        units = "μs"
    elif max_time > 1e-9:
        scalar = 1e9
        units = "ns"
    elif max_time > 1e-12:
        scalar = 1e12
        units = "ps"
    else:
        scalar = 1e15
        units = "fs"

    return units, scalar


def freq_units(freq: npt.NDArray[np.float64]) -> tuple[str, float]:
    """
    Determines the appropriate frequency units to use for a given frequency array.
    """
    max_freq = np.nanmax(np.abs(freq))

    if max_freq > 1e12:
        scalar = 1e-12
        units = "THz"
    elif max_freq > 1e9:
        scalar = 1e-9
        units = "GHz"
    elif max_freq > 1e6:
        scalar = 1e-6
        units = "MHz"
    elif max_freq > 1e3:
        scalar = 1e-3
        units = "kHz"
    elif max_freq > 1:
        scalar = 1
        units = "Hz"
    elif max_freq > 1e-3:
        scalar = 1e3
        units = "mHz"
    else:
        scalar = 1e6
        units = "μHz"

    return units, scalar



"""
MATLAB:
    for N = [1, 2, 3, 4, 5, 7, 11, 13]
        barker = comm.BarkerCode("Length", N, "SamplesPerFrame", N);
        disp(N)
        disp(out)
        disp(barker())
    end
"""

import numpy as np

import sdr


def verify_code(length, sequence_truth):
    sequence_truth = np.array(sequence_truth)

    sequence = sdr.barker_code(length, output="bipolar")
    assert isinstance(sequence, np.ndarray)
    assert np.array_equal(sequence, sequence_truth)

    code = sdr.barker_code(length, output="binary")
    assert isinstance(code, np.ndarray)
    assert np.array_equal(code, (1 - sequence_truth) // 2)


def test_1():
    verify_code(1, [-1])


def test_2():
    verify_code(2, [-1, 1])


def test_3():
    verify_code(3, [-1, -1, 1])


def test_4():
    verify_code(4, [-1, -1, 1, -1])


def test_5():
    verify_code(5, [-1, -1, -1, 1, -1])


def test_7():
    verify_code(7, [-1, -1, -1, 1, 1, -1, 1])


def test_11():
    verify_code(11, [-1, -1, -1, 1, 1, 1, -1, 1, 1, -1, 1])


def test_13():
    verify_code(13, [-1, -1, -1, -1, -1, 1, 1, -1, -1, 1, -1, 1, -1])



"""
A module for infinite impulse response (IIR) filters.
"""

from __future__ import annotations

from typing import Any, overload

import numpy as np
import numpy.typing as npt
import scipy.integrate
import scipy.signal
from typing_extensions import Self

from .._helper import convert_output, export, verify_arraylike, verify_bool, verify_scalar
from ._common import frequency_response


@export
class IIR:
    r"""
    Implements an infinite impulse response (IIR) filter.

    This class is a wrapper for the :func:`scipy.signal.lfilter` function. It supports one-time filtering
    and streamed filtering.

    Notes:
        An IIR filter is defined by its feedforward coefficients $b_i$ and feedback coefficients $a_j$.
        These coefficients define the difference equation

        $$y[n] = \frac{1}{a_0} \left( \sum_{i=0}^{M} b_i x[n-i] - \sum_{j=1}^{N} a_j y[n-j] \right) .$$

        The transfer function of the filter is

        $$H(z) = \frac{\sum\limits_{i=0}^{M} b_i z^{-i}}{\sum\limits_{j=0}^{N} a_j z^{-j}} .$$

    Examples:
        See the :ref:`iir-filters` example.

    Group:
        dsp-iir-filtering
    """

    def __init__(
        self,
        b: npt.ArrayLike,
        a: npt.ArrayLike,
        streaming: bool = False,
    ):
        """
        Creates an IIR filter with feedforward coefficients $b_i$ and feedback coefficients $a_j$.

        Arguments:
            b: The feedforward coefficients $b_i$ for $i = 0,...,M$.
            a: The feedback coefficients $a_j$ for $j = 0,...,N$.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs and outputs are
                preserved between calls to :meth:`~IIR.__call__()`.

        Examples:
            See the :ref:`iir-filters` example.
        """
        self._b_taps = verify_arraylike(b, atleast_1d=1, ndim=1)
        self._a_taps = verify_arraylike(a, atleast_1d=1, ndim=1)
        self._streaming = verify_bool(streaming)

        self._state: npt.NDArray  # The filter state. Will be updated in reset().
        self.reset()

        # Compute the zeros and poles of the transfer function
        self._zeros, self._poles, self._gain = scipy.signal.tf2zpk(self.b_taps, self.a_taps)

    ##############################################################################
    # Alternate constructors
    ##############################################################################

    @classmethod
    def ZerosPoles(
        cls,
        zeros: npt.ArrayLike,
        poles: npt.ArrayLike,
        gain: float = 1.0,
        streaming: bool = False,
    ) -> Self:
        """
        Creates an IIR filter from its zeros, poles, and gain.

        Arguments:
            zeros: The zeros of the transfer function.
            poles: The poles of the transfer function.
            gain: The gain of the transfer function.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs and outputs are
                preserved between calls to :meth:`~IIR.__call__()`.

        Examples:
            See the :ref:`iir-filters` example.
        """
        zeros = verify_arraylike(zeros, atleast_1d=1, ndim=1)
        poles = verify_arraylike(poles, atleast_1d=1, ndim=1)
        verify_scalar(gain, float=True, positive=True)
        streaming = verify_bool(streaming)

        b, a = scipy.signal.zpk2tf(zeros, poles, gain)

        return cls(b, a, streaming=streaming)

    ##############################################################################
    # Special methods
    ##############################################################################

    def __call__(self, x: npt.ArrayLike) -> npt.NDArray:
        r"""
        Filters the input signal $x[n]$ with the IIR filter.

        Arguments:
            x: The input signal $x[n]$ with length $L$.

        Returns:
            The filtered signal $y[n]$ with length $L$.

        Examples:
            See the :ref:`iir-filters` example.
        """
        x = verify_arraylike(x, atleast_1d=1, ndim=1)

        if not self.streaming:
            self.reset()

        y, self._state = scipy.signal.lfilter(self.b_taps, self.a_taps, x, zi=self._state)

        return convert_output(y)

    def __repr__(self) -> str:
        return f"sdr.{type(self).__name__}({self.b_taps.tolist()}, {self.a_taps.tolist()}, streaming={self.streaming})"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  b_taps: {self.b_taps.shape} shape"
        string += f"\n    {self.b_taps.tolist()}"
        string += f"\n  a_taps: {self.a_taps.shape} shape"
        string += f"\n    {self.a_taps.tolist()}"
        string += f"\n  zeros: {self.zeros.shape} shape"
        string += f"\n    {self.zeros.tolist()}"
        string += f"\n  poles: {self.poles.shape} shape"
        string += f"\n    {self.poles.tolist()}"
        string += f"\n  streaming: {self.streaming}"
        return string

    ##############################################################################
    # Streaming mode
    ##############################################################################

    def reset(self):
        """
        Resets the filter state. Only useful when using streaming mode.

        Examples:
            See the :ref:`iir-filters` example.

        Group:
            Streaming mode only
        """
        self._state = scipy.signal.lfiltic(self.b_taps, self.a_taps, y=[], x=[])

    @property
    def streaming(self) -> bool:
        """
        Indicates whether the filter is in streaming mode.

        In streaming mode, the filter state is preserved between calls to :meth:`~IIR.__call__()`.

        Examples:
            See the :ref:`iir-filters` example.

        Group:
            Streaming mode only
        """
        return self._streaming

    @property
    def state(self) -> npt.NDArray:
        """
        The filter state.

        Examples:
            See the :ref:`iir-filters` example.

        Group:
            Streaming mode only
        """
        return self._state

    ##############################################################################
    # Methods
    ##############################################################################

    def impulse_response(self, N: int = 100) -> npt.NDArray:
        r"""
        Returns the impulse response $h[n]$ of the IIR filter.

        The impulse response $h[n]$ is the filter output when the input is an impulse $\delta[n]$.

        Arguments:
            N: The number of samples to return.

        Returns:
            The impulse response of the IIR filter $h[n]$.

        See Also:
            sdr.plot.impulse_response

        Examples:
            See the :ref:`iir-filters` example.
        """
        verify_scalar(N, int=True, positive=True)

        # Delta impulse function
        d = np.zeros(N, dtype=float)
        d[0] = 1

        zi = self._state
        h = self(d)
        self._state = zi  # Restore the filter state

        return h

    def step_response(self, N: int = 100) -> npt.NDArray:
        """
        Returns the step response $s[n]$ of the IIR filter.

        The step response $s[n]$ is the filter output when the input is a unit step $u[n]$.

        Arguments:
            N: The number of samples to return.

        Returns:
            The step response of the IIR filter $s[n]$.

        See Also:
            sdr.plot.step_response

        Examples:
            See the :ref:`iir-filters` example.
        """
        verify_scalar(N, int=True, positive=True)

        # Unit step function
        u = np.ones(N, dtype=float)

        state = self._state
        s = self(u)
        self._state = state  # Restore the filter state

        return s

    @overload
    def frequency_response(
        self,
        freqs: int = 1024,
        sample_rate: float = 1.0,
        whole: bool = True,
        decades: int | None = None,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.complex128]]: ...

    @overload
    def frequency_response(
        self,
        freqs: float | npt.NDArray[np.float64],
        sample_rate: float = 1.0,
    ) -> complex: ...

    @overload
    def frequency_response(
        self,
        freqs: npt.NDArray[np.float64],
        sample_rate: float = 1.0,
    ) -> npt.NDArray[np.complex128]: ...

    def frequency_response(
        self,
        freqs: Any = 1024,
        sample_rate: Any = 1.0,
        whole: Any = True,
        decades: Any | None = None,
    ) -> Any:
        r"""
        Returns the frequency response $H(\omega)$ of the IIR filter.

        Arguments:
            freqs: The frequency specification.

                - `int`: The number of frequency points. The endpoint is not included.
                - `float`: A single frequency.
                - `npt.NDArray[float]`: Multiple frequencies.

            sample_rate: The sample rate $f_s$ of the filter in samples/s.
            whole: Only used if `freqs` is an integer.

                - `True`: The maximum frequency is `max_f = sample_rate`.
                - `False`: The maximum frequency is `max_f = sample_rate / 2`.

            decades: Only used if `freqs` is an integer.

                - `None`: `f = np.linspace(0, max_f, freqs, endpoint=False)`.
                - `int`: `f = np.logspace(np.log10(max_f) - decades), np.log10(max_f), freqs, endpoint=False)`.

        Returns:
            - The frequency vector $f$, only if `freqs` is an integer.
            - The frequency response of the IIR filter $H(\omega)$.

        See Also:
            sdr.plot.magnitude_response, sdr.plot.phase_response

        Examples:
            .. ipython:: python

                zero = 0.6; \
                pole = 0.8 * np.exp(1j * np.pi / 8); \
                iir = sdr.IIR.ZerosPoles([zero], [pole, pole.conj()])

            Compute the frequency response at 1024 evenly spaced frequencies.

            .. ipython:: python

                iir.frequency_response()

            Compute the frequency response at 0.0 rad/s.

            .. ipython:: python

                iir.frequency_response(0.0)

            Compute the frequency response at several frequencies in Hz.

            .. ipython:: python

                iir.frequency_response([100, 200, 300, 400], sample_rate=1000)
        """
        verify_scalar(sample_rate, float=True, positive=True)
        verify_bool(whole)
        verify_scalar(decades, optional=True, int=True, positive=True)

        f, H = frequency_response(self.b_taps, self.a_taps, freqs, sample_rate, whole, decades)
        if isinstance(freqs, int):
            return f, H
        elif isinstance(freqs, float):
            return H[0]
        else:
            return H

    def noise_bandwidth(self, sample_rate: float = 1.0) -> float:
        r"""
        Returns the noise bandwidth $B_n$ of the IIR filter.

        Arguments:
            sample_rate: The sample rate $f_s$ of the filter in samples/s.

        Returns:
            The noise bandwidth of the IIR filter $B_n$ in Hz.

        Examples:
            See the :ref:`iir-filters` example.
        """
        verify_scalar(sample_rate, float=True, positive=True)

        # Compute the frequency response
        f, H = scipy.signal.freqz(self.b_taps, self.a_taps, worN=2**20, whole=True, fs=sample_rate)
        H = np.abs(H) ** 2

        # Compute the noise bandwidth
        B_n = scipy.integrate.simpson(y=H, x=f) / np.max(H)

        return B_n

    ##############################################################################
    # Properties
    ##############################################################################

    @property
    def b_taps(self) -> npt.NDArray:
        """
        The feedforward taps $b_i$ for $i = 0,...,M$.

        Examples:
            See the :ref:`iir-filters` example.
        """
        return self._b_taps

    @property
    def a_taps(self) -> npt.NDArray:
        """
        The feedback taps $a_j$ for $j = 0,...,N$.

        Examples:
            See the :ref:`iir-filters` example.
        """
        return self._a_taps

    @property
    def order(self) -> int:
        """
        The order of the IIR filter $N$.

        Examples:
            See the :ref:`iir-filters` example.
        """
        return self._a_taps.size - 1

    @property
    def zeros(self) -> npt.NDArray:
        """
        The zeros of the IIR filter.

        Examples:
            See the :ref:`iir-filters` example.
        """
        return self._zeros

    @property
    def poles(self) -> npt.NDArray:
        """
        The poles of the IIR filter.

        Examples:
            See the :ref:`iir-filters` example.
        """
        return self._poles

    @property
    def gain(self) -> float:
        """
        The gain of the IIR filter.

        Examples:
            See the :ref:`iir-filters` example.
        """
        return self._gain



"""
A pytest module to test the Fibonacci LFSR implementation.
"""

import galois
import numpy as np
import pytest

import sdr

# These two polys allow for testing JIT compiled and pure-Python implementations
CHARACTERISTIC_POLYS = [
    galois.primitive_poly(7, 4),
    galois.Poly.Str(
        "x^4 + 414029366129716807589746234643x^3 + 713840634647528950143955598853x^2 + 178965232760409569156590479285x + 574717025925479275195710910921",
        field=galois.GF(2**100),
    ),
]


def test_from_taps():
    GF = galois.GF(7)
    T = GF([1, 2, 3, 4])
    lfsr = sdr.FLFSR.Taps(T)
    assert lfsr.characteristic_poly == galois.Poly([1, -1, -2, -3, -4], field=GF)
    assert lfsr.feedback_poly == galois.Poly([-4, -3, -2, -1, 1], field=GF)


def test_repr():
    c = galois.primitive_poly(7, 4)
    lfsr = sdr.FLFSR(c)
    assert repr(lfsr) == "<Fibonacci LFSR: c(x) = x^4 + x^2 + 3x + 5 over GF(7)>"


def test_str():
    c = galois.primitive_poly(7, 4)
    lfsr = sdr.FLFSR(c)
    assert (
        str(lfsr)
        == "Fibonacci LFSR:\n  field: GF(7)\n  characteristic_poly: x^4 + x^2 + 3x + 5\n  feedback_poly: 5x^4 + 3x^3 + x^2 + 1\n  taps: [0 6 4 2]\n  order: 4\n  state: [1 1 1 1]\n  initial_state: [1 1 1 1]"
    )


@pytest.mark.parametrize("c", CHARACTERISTIC_POLYS)
def test_initial_state(c):
    default_state = [1, 1, 1, 1]
    lfsr = sdr.FLFSR(c)
    assert np.array_equal(lfsr.initial_state, default_state)
    assert np.array_equal(lfsr.state, default_state)

    state = [1, 2, 3, 4]
    lfsr = sdr.FLFSR(c, state=state)
    assert np.array_equal(lfsr.initial_state, state)
    assert np.array_equal(lfsr.state, state)


@pytest.mark.parametrize("c", CHARACTERISTIC_POLYS)
def test_feedback_and_characteristic_poly(c):
    lfsr = sdr.FLFSR(c)
    assert lfsr.feedback_poly == c.reverse()
    assert lfsr.characteristic_poly == c
    assert lfsr.feedback_poly == lfsr.characteristic_poly.reverse()


def test_reset_exceptions():
    c = galois.primitive_poly(7, 4)
    lfsr = sdr.FLFSR(c)

    with pytest.raises(TypeError):
        lfsr.reset(1)


@pytest.mark.parametrize("c", CHARACTERISTIC_POLYS)
def test_reset_initial_state(c):
    lfsr = sdr.FLFSR(c)

    assert np.array_equal(lfsr.state, lfsr.initial_state)
    lfsr.step(10)
    assert not np.array_equal(lfsr.state, lfsr.initial_state)
    lfsr.reset()
    assert np.array_equal(lfsr.state, lfsr.initial_state)


@pytest.mark.parametrize("c", CHARACTERISTIC_POLYS)
def test_reset_specific_state(c):
    lfsr = sdr.FLFSR(c)
    state = [1, 2, 3, 4]

    assert not np.array_equal(lfsr.state, state)
    lfsr.reset(state)
    assert np.array_equal(lfsr.state, state)


def test_step_exceptions():
    c = galois.primitive_poly(7, 4)
    lfsr = sdr.FLFSR(c)

    with pytest.raises(TypeError):
        lfsr.step(10.0)


@pytest.mark.parametrize("c", CHARACTERISTIC_POLYS)
def test_step_zero(c):
    lfsr = sdr.FLFSR(c)

    y = lfsr.step(0)
    assert y.size == 0
    assert type(y) is lfsr.field


@pytest.mark.parametrize("c", CHARACTERISTIC_POLYS)
def test_step_forwards_backwards(c):
    lfsr = sdr.FLFSR(c)

    y_forward = lfsr.step(10)
    y_reverse = lfsr.step(-10)

    assert np.array_equal(y_forward, y_reverse[::-1])
    assert np.array_equal(lfsr.state, lfsr.initial_state)

    # Step forward and back by 1 step
    assert lfsr.step(1) == lfsr.step(-1)
    assert lfsr.step(1) == lfsr.step(-1)
    assert lfsr.step(1) == lfsr.step(-1)
    assert np.array_equal(lfsr.state, lfsr.initial_state)


@pytest.mark.parametrize("c", CHARACTERISTIC_POLYS)
def test_step_backwards_forwards(c):
    lfsr = sdr.FLFSR(c)

    y_reverse = lfsr.step(-10)
    y_forward = lfsr.step(10)
    assert np.array_equal(y_reverse, y_forward[::-1])
    assert np.array_equal(lfsr.state, lfsr.initial_state)

    # Step backward and forward by 1 step
    assert lfsr.step(-1) == lfsr.step(1)
    assert lfsr.step(-1) == lfsr.step(1)
    assert lfsr.step(-1) == lfsr.step(1)
    assert np.array_equal(lfsr.state, lfsr.initial_state)


@pytest.mark.parametrize("c", CHARACTERISTIC_POLYS)
def test_step_output_reversed_state(c):
    state = [1, 2, 3, 4]
    lfsr = sdr.FLFSR(c, state=state)

    y = lfsr.step(4)
    assert np.array_equal(y, state[::-1])


def test_step_gf2_primitive():
    """
    Python:
        c = galois.primitive_poly(2, 4); c
        key = -c.coeffs[1:][::-1]; key

    Sage:
        F = GF(2)
        key = [1, 1, 0, 0]  # c(x) = x^4 + x + 1
        fill = [1, 1, 1, 1]
        key = [F(k) for k in key]
        fill = [F(f) for f in fill]
        s = lfsr_sequence(key, fill, 50)
        print(s)
    """
    GF = galois.GF(2)
    key = GF([1, 1, 0, 0])
    key = -key[::-1]
    key = np.insert(key, 0, 1)  # Convert to c(x)
    c = galois.Poly(key)
    state = GF([1, 1, 1, 1])
    y_truth = GF([1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 1, 1, 1, 0])  # fmt: skip

    lfsr = sdr.FLFSR(c, state=state)
    y = lfsr.step(50)

    assert np.array_equal(y, y_truth)
    assert type(y) is lfsr.field


def test_step_gf3_primitive():
    """
    Python:
        c = galois.primitive_poly(3, 4); c
        key = -c.coeffs[1:][::-1]; key

    Sage:
        F = GF(3)
        key = [1, 2, 0, 0]  # c(x) = x^4 + x + 2
        fill = [1, 1, 1, 1]
        key = [F(k) for k in key]
        fill = [F(f) for f in fill]
        s = lfsr_sequence(key, fill, 50)
        print(s)
    """
    GF = galois.GF(3)
    key = GF([1, 2, 0, 0])
    key = -key[::-1]
    key = np.insert(key, 0, 1)  # Convert to c(x)
    c = galois.Poly(key)
    state = GF([1, 1, 1, 1])
    y_truth = GF([1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 2, 1, 0, 1, 1, 1, 2, 0, 0, 2, 2, 0, 1, 0, 2, 2, 1, 1, 0, 1, 0, 1, 2, 1, 2, 2, 1, 2, 0, 1, 2, 2, 2, 2, 0, 0, 0, 2, 0, 0])  # fmt: skip

    lfsr = sdr.FLFSR(c, state=state)
    y = lfsr.step(50)

    assert np.array_equal(y, y_truth)
    assert type(y) is lfsr.field


def test_step_gf2_3_primitive():
    """
    Python:
        c = galois.primitive_poly(2**3, 4); c
        key = -c.coeffs[1:][::-1]; key

    Sage:
        F = GF(2^3, repr="int")
        key = [3, 1, 0, 0]  # c(x) = x^4 + x + 3
        fill = [1, 1, 1, 1]
        key = [F.fetch_int(k) for k in key]
        fill = [F.fetch_int(f) for f in fill]
        s = lfsr_sequence(key, fill, 50)
        print(s)
    """
    GF = galois.GF(2**3)
    key = GF([3, 1, 0, 0])
    key = -key[::-1]
    key = np.insert(key, 0, 1)  # Convert to c(x)
    c = galois.Poly(key)
    state = GF([1, 1, 1, 1])
    y_truth = GF([1, 1, 1, 1, 2, 2, 2, 1, 4, 4, 7, 7, 3, 0, 5, 1, 5, 5, 5, 6, 1, 1, 2, 0, 2, 1, 6, 2, 7, 5, 3, 1, 7, 7, 4, 4, 5, 6, 3, 2, 2, 2, 7, 4, 4, 1, 6, 3, 6, 5])  # fmt: skip

    lfsr = sdr.FLFSR(c, state=state)
    y = lfsr.step(50)

    assert np.array_equal(y, y_truth)
    assert type(y) is lfsr.field


def test_step_gf3_3_primitive():
    """
    Python:
        c = galois.primitive_poly(3**3, 4); c
        key = -c.coeffs[1:][::-1]; key

    Sage:
        F = GF(3^3, repr="int")
        key = [20, 2, 0, 0]  # c(x) = x^4 + x + 10
        fill = [1, 1, 1, 1]
        key = [F.fetch_int(k) for k in key]
        fill = [F.fetch_int(f) for f in fill]
        s = lfsr_sequence(key, fill, 50)
        print(s)
    """
    GF = galois.GF(3**3)
    key = GF([20, 2, 0, 0])
    key = -key[::-1]
    key = np.insert(key, 0, 1)  # Convert to c(x)
    c = galois.Poly(key)
    state = GF([1, 1, 1, 1])
    y_truth = GF([1, 1, 1, 1, 19, 19, 19, 1, 25, 25, 16, 4, 24, 6, 6, 6, 26, 2, 2, 9, 4, 11, 1, 11, 13, 21, 9, 9, 12, 10, 3, 0, 6, 2, 4, 3, 6, 15, 18, 7, 20, 20, 20, 8, 17, 17, 2, 1, 13, 19])  # fmt: skip

    lfsr = sdr.FLFSR(c, state=state)
    y = lfsr.step(50)

    assert np.array_equal(y, y_truth)
    assert type(y) is lfsr.field


def test_step_gf2_reducible():
    """
    Python:
        GF = galois.GF(2)
        coeffs = GF.Random(5)
        coeffs[0] = 1
        coeffs[-1] = GF.Random(low=1)
        c = galois.Poly(coeffs)
        assert not c.is_irreducible()
        key = -coeffs[1:][::-1]
        c
        key

    Sage:
        F = GF(2)
        key = [1, 0, 0, 0]  # c(x) = x^4 + 1
        fill = [1, 1, 1, 1]
        key = [F(k) for k in key]
        fill = [F(f) for f in fill]
        s = lfsr_sequence(key, fill, 50)
        print(s)
    """
    GF = galois.GF(2)
    key = GF([1, 0, 0, 0])
    key = -key[::-1]
    key = np.insert(key, 0, 1)  # Convert to c(x)
    c = galois.Poly(key)
    state = GF([1, 1, 1, 1])
    y_truth = GF([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])  # fmt: skip

    lfsr = sdr.FLFSR(c, state=state)
    y = lfsr.step(50)

    assert np.array_equal(y, y_truth)
    assert type(y) is lfsr.field


def test_step_gf3_reducible():
    """
    Python:
        GF = galois.GF(3)
        coeffs = GF.Random(5)
        coeffs[0] = 1
        coeffs[-1] = GF.Random(low=1)
        c = galois.Poly(coeffs)
        assert not c.is_irreducible()
        key = -coeffs[1:][::-1]
        c
        key

    Sage:
        F = GF(3)
        key = [1, 2, 2, 1]  # c(x) = x^4 + x^3 + 2x^2 + 2x + 1
        fill = [1, 1, 1, 1]
        key = [F(k) for k in key]
        fill = [F(f) for f in fill]
        s = lfsr_sequence(key, fill, 50)
        print(s)
    """
    GF = galois.GF(3)
    key = GF([1, 2, 2, 1])
    key = -key[::-1]
    key = np.insert(key, 0, 1)  # Convert to c(x)
    c = galois.Poly(key)
    state = GF([1, 1, 1, 1])
    y_truth = GF([1, 1, 1, 1, 0, 2, 2, 1, 0, 2, 0, 2, 0, 0, 1, 0, 2, 1, 0, 0, 1, 2, 1, 1, 2, 2, 0, 0, 0, 2, 2, 0, 2, 2, 2, 1, 2, 1, 0, 1, 2, 2, 2, 2, 0, 1, 1, 2, 0, 1])  # fmt: skip

    lfsr = sdr.FLFSR(c, state=state)
    y = lfsr.step(50)

    assert np.array_equal(y, y_truth)
    assert type(y) is lfsr.field


def test_step_gf2_3_reducible():
    """
    Python:
        GF = galois.GF(2**3)
        coeffs = GF.Random(5)
        coeffs[0] = 1
        coeffs[-1] = GF.Random(low=1)
        c = galois.Poly(coeffs)
        assert not c.is_irreducible()
        key = -coeffs[1:][::-1]
        c
        key

    Sage:
        F = GF(2^3, repr="int")
        key = [4, 1, 1, 5]  # c(x) = x^4 + 5x^3 + x^2 + x + 4
        fill = [1, 1, 1, 1]
        key = [F.fetch_int(k) for k in key]
        fill = [F.fetch_int(f) for f in fill]
        s = lfsr_sequence(key, fill, 50)
        print(s)
    """
    GF = galois.GF(2**3)
    key = GF([4, 1, 1, 5])
    key = -key[::-1]
    key = np.insert(key, 0, 1)  # Convert to c(x)
    c = galois.Poly(key)
    state = GF([1, 1, 1, 1])
    y_truth = GF([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])  # fmt: skip

    lfsr = sdr.FLFSR(c, state=state)
    y = lfsr.step(50)

    assert np.array_equal(y, y_truth)
    assert type(y) is lfsr.field


def test_step_gf3_3_reducible():
    """
    Python:
        GF = galois.GF(3**3)
        coeffs = GF.Random(5)
        coeffs[0] = 1
        coeffs[-1] = GF.Random(low=1)
        c = galois.Poly(coeffs)
        assert not c.is_irreducible()
        key = -coeffs[1:][::-1]
        c
        key

    Sage:
        F = GF(3^3, repr="int")
        key = [5, 20, 23, 5]  # c(x) = x^4 + 5x^3 + 23x^2 + 20x + 5
        fill = [1, 1, 1, 1]
        key = [F.fetch_int(k) for k in key]
        fill = [F.fetch_int(f) for f in fill]
        s = lfsr_sequence(key, fill, 50)
        print(s)
    """
    GF = galois.GF(3**3)
    key = GF([5, 20, 23, 5])
    key = -key[::-1]
    key = np.insert(key, 0, 1)  # Convert to c(x)
    c = galois.Poly(key)
    state = GF([1, 1, 1, 1])
    y_truth = GF([1, 1, 1, 1, 11, 6, 1, 16, 20, 13, 6, 13, 2, 9, 18, 8, 21, 6, 12, 6, 3, 3, 26, 7, 22, 16, 23, 13, 5, 6, 1, 7, 19, 3, 12, 16, 14, 16, 17, 6, 0, 24, 9, 26, 6, 23, 3, 22, 21, 8])  # fmt: skip

    lfsr = sdr.FLFSR(c, state=state)
    y = lfsr.step(50)

    assert np.array_equal(y, y_truth)
    assert type(y) is lfsr.field


@pytest.mark.parametrize("c", CHARACTERISTIC_POLYS)
def test_to_galois_lfsr(c):
    fibonacci_lfsr = sdr.FLFSR(c)
    galois_lfsr = fibonacci_lfsr.to_galois_lfsr()

    y1 = fibonacci_lfsr.step(100)
    y2 = galois_lfsr.step(100)

    assert np.array_equal(y1, y2)


@pytest.mark.parametrize("order", [2, 3, 2**3, 3**3])
def test_to_galois_lfsr_primitive(order):
    c = galois.primitive_poly(order, 4)

    fibonacci_lfsr = sdr.FLFSR(c)
    galois_lfsr = fibonacci_lfsr.to_galois_lfsr()

    y1 = fibonacci_lfsr.step(100)
    y2 = galois_lfsr.step(100)

    assert np.array_equal(y1, y2)


@pytest.mark.parametrize("order", [2, 3, 2**3, 3**3])
def test_to_galois_lfsr_reducible(order):
    GF = galois.GF(order)
    while True:
        coeffs = GF.Random(5)
        coeffs[0] = 1
        coeffs[-1] = GF.Random(low=1)
        c = galois.Poly(coeffs)
        if not c.is_irreducible():
            break

    fibonacci_lfsr = sdr.FLFSR(c)
    galois_lfsr = fibonacci_lfsr.to_galois_lfsr()

    y1 = fibonacci_lfsr.step(100)
    y2 = galois_lfsr.step(100)

    assert np.array_equal(y1, y2)



"""
A module containing various probability functions.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.special

from ._helper import convert_output, export, verify_arraylike, verify_scalar


@export
def Q(x: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Computes the CCDF of the standard normal distribution $\mathcal{N}(0, 1)$..

    The complementary cumulative distribution function (CCDF) $Q(x)$ is the probability that a random variable
    exceeds a given value.

    Arguments:
        x: The real-valued input $x$.

    Returns:
        The probability $p$ that $x$ is exceeded.

    See Also:
        sdr.Qinv

    Examples:
        .. ipython:: python

            sdr.Q(1)
            sdr.Qinv(0.158655)

    Group:
        probability
    """
    x = verify_arraylike(x, float=True)

    p = scipy.special.erfc(x / np.sqrt(2)) / 2

    return convert_output(p)


@export
def Qinv(p: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Computes the inverse CCDF of the standard normal distribution $\mathcal{N}(0, 1)$.

    The inverse complementary cumulative distribution function (CCDF) $Q^{-1}(p)$ is the value that is exceeded
    with a given probability.

    Arguments:
        p: The probability $p$ of exceeding the returned value $x$.

    Returns:
        The real-valued $x$ that is exceeded with probability $p$.

    See Also:
        sdr.Q

    Examples:
        .. ipython:: python

            sdr.Qinv(0.158655)
            sdr.Q(1)

    Group:
        probability
    """
    p = verify_arraylike(p, float=True)

    x = np.sqrt(2) * scipy.special.erfcinv(2 * p)

    return convert_output(x)


@export
def sum_distribution(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    n_terms: int,
    p: float = 1e-16,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the sum of $n$ i.i.d. random variables $X_i$.

    Arguments:
        X: The distribution of the i.i.d. random variables $X_i$.
        n_terms: The number $n$ of random variables to sum.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the sum $Z = X_1 + X_2 + \ldots + X_n$.

    Notes:
        The PDF of the sum of $n$ independent random variables is the convolution of the PDF of the base distribution.

        $$f_{X_1 + X_2 + \ldots + X_n}(t) = (f_X * f_X * \ldots * f_X)(t)$$

    Examples:
        Compute the distribution of the sum of two normal distributions.

        .. ipython:: python

            X = scipy.stats.norm(loc=-1, scale=0.5)
            n_terms = 2
            x = np.linspace(-6, 2, 1_001)

            @savefig sdr_sum_distribution_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="X"); \
            plt.plot(x, sdr.sum_distribution(X, n_terms).pdf(x), label="X + X"); \
            plt.hist(X.rvs((100_000, n_terms)).sum(axis=1), bins=101, density=True, histtype="step", label="X + X empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of two Normal distributions");

        Compute the distribution of the sum of three Rayleigh distributions.

        .. ipython:: python

            X = scipy.stats.rayleigh(scale=1)
            n_terms = 3
            x = np.linspace(0, 10, 1_001)

            @savefig sdr_sum_distribution_2.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="X"); \
            plt.plot(x, sdr.sum_distribution(X, n_terms).pdf(x), label="X + X + X"); \
            plt.hist(X.rvs((100_000, n_terms)).sum(axis=1), bins=101, density=True, histtype="step", label="X + X + X empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of three Rayleigh distributions");

        Compute the distribution of the sum of four Rician distributions.

        .. ipython:: python

            X = scipy.stats.rice(2)
            n_terms = 4
            x = np.linspace(0, 18, 1_001)

            @savefig sdr_sum_distribution_3.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="X"); \
            plt.plot(x, sdr.sum_distribution(X, n_terms).pdf(x), label="X + X + X + X"); \
            plt.hist(X.rvs((100_000, n_terms)).sum(axis=1), bins=101, density=True, histtype="step", label="X + X + X + X empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of four Rician distributions");

    Group:
        probability
    """
    verify_scalar(n_terms, int=True, positive=True)
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    if n_terms == 1:
        return X

    # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
    # is p.
    x1_min, x1_max = _x_range(X, p)
    x = np.linspace(x1_min, x1_max, 1_001)
    dx = np.mean(np.diff(x))

    # Compute the PDF of the base distribution
    f_X = X.pdf(x)

    # The PDF of the sum of n_terms independent random variables is the convolution of the PDF of the base distribution.
    # This is efficiently computed in the frequency domain by exponentiating the FFT. The FFT must be zero-padded
    # enough that the circular convolutions do not wrap around.
    n_fft = scipy.fft.next_fast_len(f_X.size * n_terms)
    f_X_fft = np.fft.fft(f_X, n_fft)
    f_X_fft = f_X_fft**n_terms
    f_Y = np.fft.ifft(f_X_fft).real
    f_Y /= f_Y.sum() * dx
    x = np.arange(f_Y.size) * dx + x[0] * (n_terms)

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    x = np.append(x, x[-1] + dx)
    x -= dx / 2

    return scipy.stats.rv_histogram((f_Y, x))


@export
def sum_distributions(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    Y: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    p: float = 1e-16,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the sum of two independent random variables $X$ and $Y$.

    Arguments:
        X: The distribution of the first random variable $X$.
        Y: The distribution of the second random variable $Y$.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the sum $Z = X + Y$.

    Notes:
        The PDF of the sum of two independent random variables is the convolution of the PDF of the two distributions.

        $$f_{X+Y}(t) = (f_X * f_Y)(t)$$

    Examples:
        Compute the distribution of the sum of two normal distributions.

        .. ipython:: python

            X = scipy.stats.norm(loc=-1, scale=0.5)
            Y = scipy.stats.norm(loc=2, scale=1.5)
            x = np.linspace(-5, 10, 1_001)

            @savefig sdr_sum_distributions_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="X"); \
            plt.plot(x, Y.pdf(x), label="Y"); \
            plt.plot(x, sdr.sum_distributions(X, Y).pdf(x), label="X + Y"); \
            plt.hist(X.rvs(100_000) + Y.rvs(100_000), bins=101, density=True, histtype="step", label="X + Y empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of two Normal distributions");

        Compute the distribution of the sum of two Rayleigh distributions.

        .. ipython:: python

            X = scipy.stats.rayleigh(scale=1)
            Y = scipy.stats.rayleigh(loc=1, scale=2)
            x = np.linspace(0, 12, 1_001)

            @savefig sdr_sum_distributions_2.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="X"); \
            plt.plot(x, Y.pdf(x), label="Y"); \
            plt.plot(x, sdr.sum_distributions(X, Y).pdf(x), label="X + Y"); \
            plt.hist(X.rvs(100_000) + Y.rvs(100_000), bins=101, density=True, histtype="step", label="X + Y empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of two Rayleigh distributions");

        Compute the distribution of the sum of two Rician distributions.

        .. ipython:: python

            X = scipy.stats.rice(2)
            Y = scipy.stats.rice(3)
            x = np.linspace(0, 12, 1_001)

            @savefig sdr_sum_distributions_3.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="X"); \
            plt.plot(x, Y.pdf(x), label="Y"); \
            plt.plot(x, sdr.sum_distributions(X, Y).pdf(x), label="X + Y"); \
            plt.hist(X.rvs(100_000) + Y.rvs(100_000), bins=101, density=True, histtype="step", label="X + Y empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Sum of two Rician distributions");

    Group:
        probability
    """
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
    # is p.
    x1_min, x1_max = _x_range(X, p)
    x2_min, x2_max = _x_range(Y, p)
    dx1 = (x1_max - x1_min) / 1_000
    dx2 = (x2_max - x2_min) / 1_000
    dx = np.min([dx1, dx2])  # Use the smaller delta x -- must use the same dx for both distributions
    x1 = np.arange(x1_min, x1_max, dx)
    x2 = np.arange(x2_min, x2_max, dx)

    # Compute the PDF of each distribution
    f_X = X.pdf(x1)
    f_Y = Y.pdf(x2)

    # The PDF of the sum of two independent random variables is the convolution of the PDF of the two distributions
    f_Z = np.convolve(f_X, f_Y, mode="full") * dx

    # Determine the x axis for the output convolution
    x = np.arange(f_Z.size) * dx + x1[0] + x2[0]

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    x = np.append(x, x[-1] + dx)
    x -= dx / 2

    return scipy.stats.rv_histogram((f_Z, x))


@export
def multiply_distributions(
    X: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    Y: scipy.stats.rv_continuous | scipy.stats.rv_histogram,
    x: npt.ArrayLike | None = None,
    p: float = 1e-10,
) -> scipy.stats.rv_histogram:
    r"""
    Numerically calculates the distribution of the product of two independent random variables $X$ and $Y$.

    Arguments:
        X: The distribution of the first random variable $X$.
        Y: The distribution of the second random variable $Y$.
        x: The x values at which to evaluate the PDF of the product. If None, the x values are determined based on `p`.
        p: The probability of exceeding the x axis, on either side, for each distribution. This is used to determine
            the bounds on the x axis for the numerical convolution. Smaller values of $p$ will result in more accurate
            analysis, but will require more computation.

    Returns:
        The distribution of the product $Z = X \cdot Y$.

    Notes:
        The PDF of the product of two independent random variables is calculated as follows.

        $$
        f_{X \cdot Y}(t) =
        \int_{0}^{\infty} f_X(k) f_Y(t/k) \frac{1}{k} dk -
        \int_{-\infty}^{0} f_X(k) f_Y(t/k) \frac{1}{k} dk
        $$

    Examples:
        Compute the distribution of the product of two normal distributions.

        .. ipython:: python

            X = scipy.stats.norm(loc=-1, scale=0.5)
            Y = scipy.stats.norm(loc=2, scale=1.5)
            x = np.linspace(-15, 10, 1_001)

            @savefig sdr_multiply_distributions_1.png
            plt.figure(); \
            plt.plot(x, X.pdf(x), label="X"); \
            plt.plot(x, Y.pdf(x), label="Y"); \
            plt.plot(x, sdr.multiply_distributions(X, Y).pdf(x), label="X * Y"); \
            plt.hist(X.rvs(100_000) * Y.rvs(100_000), bins=101, density=True, histtype="step", label="X * Y empirical"); \
            plt.legend(); \
            plt.xlabel("Random variable"); \
            plt.ylabel("Probability density"); \
            plt.title("Product of two Normal distributions");

    Group:
        probability
    """
    verify_scalar(p, float=True, exclusive_min=0, inclusive_max=0.1)

    if x is None:
        # Determine the x axis of each distribution such that the probability of exceeding the x axis, on either side,
        # is p.
        x1_min, x1_max = _x_range(X, np.sqrt(p))
        x2_min, x2_max = _x_range(Y, np.sqrt(p))
        bounds = np.array([x1_min * x2_min, x1_min * x2_max, x1_max * x2_min, x1_max * x2_max])
        x_min = np.min(bounds)
        x_max = np.max(bounds)
        x = np.linspace(x_min, x_max, 1_001)
    else:
        x = verify_arraylike(x, float=True, atleast_1d=True, ndim=1)
        x = np.sort(x)
    dx = np.mean(np.diff(x))

    def integrand(k: float, xi: float) -> float:
        return X.pdf(k) * Y.pdf(xi / k) * 1 / k

    f_Z = np.zeros_like(x)
    for i, xi in enumerate(x):
        f_Z[i] = scipy.integrate.quad(integrand, 0, np.inf, args=(xi,))[0]
        f_Z[i] -= scipy.integrate.quad(integrand, -np.inf, 0, args=(xi,))[0]

    # Adjust the histograms bins to be on either side of each point. So there is one extra point added.
    x = np.append(x, x[-1] + dx)
    x -= dx / 2

    return scipy.stats.rv_histogram((f_Z, x))


def _x_range(X: scipy.stats.rv_continuous, p: float) -> tuple[float, float]:
    r"""
    Determines the range of x values for a given distribution such that the probability of exceeding the x axis, on
    either side, is p.
    """
    # Need to have these loops because for very small p, sometimes SciPy will return NaN instead of a valid value.
    # The while loops will increase the p value until a valid value is returned.

    pp = p
    while True:
        x_min = X.ppf(pp)
        if not np.isnan(x_min):
            break
        pp *= 10

    pp = p
    while True:
        x_max = X.isf(pp)
        if not np.isnan(x_max):
            break
        pp *= 10

    return x_min, x_max



import numpy as np
import pytest

import sdr


def test_scalar_snr_in():
    assert sdr.non_coherent_gain(1, 10, snr_ref="input") == pytest.approx(0.0)
    assert sdr.non_coherent_gain(2, 10, snr_ref="input") == pytest.approx(2.499445060713011)
    assert sdr.non_coherent_gain(10, 10, snr_ref="input") == pytest.approx(8.666092814306324)
    assert sdr.non_coherent_gain(20, 10, snr_ref="input") == pytest.approx(11.410342926869486)


def test_scalar_snr_out():
    assert sdr.non_coherent_gain(1, 10, snr_ref="output") == pytest.approx(0.0)
    assert sdr.non_coherent_gain(2, 10, snr_ref="output") == pytest.approx(2.3760378084124643)
    assert sdr.non_coherent_gain(10, 10, snr_ref="output") == pytest.approx(7.367080094838606)
    assert sdr.non_coherent_gain(20, 10, snr_ref="output") == pytest.approx(9.296533509663906)


def test_vector_snr_in():
    snr = 10
    n_nc = np.array([1, 2, 10, 20])
    g_nc = sdr.non_coherent_gain(n_nc, snr, snr_ref="input")
    g_nc_truth = np.array([0.0, 2.499445060713011, 8.666092814306324, 11.410342926869486])
    assert isinstance(g_nc, np.ndarray)
    assert np.allclose(g_nc, g_nc_truth)


def test_vector_snr_out():
    snr = 10
    n_nc = np.array([1, 2, 10, 20])
    g_nc = sdr.non_coherent_gain(n_nc, snr, snr_ref="output")
    g_nc_truth = np.array([0.0, 2.3760378084124643, 7.367080094838606, 9.296533509663906])
    assert isinstance(g_nc, np.ndarray)
    assert np.allclose(g_nc, g_nc_truth)



"""
The equation was compared against plots from a technical paper. These values were then calculated once using the
"verified" code.
"""

import sdr


def test_0p5():
    p_d = 0.5
    p_fa = 1e-2
    assert sdr.peebles(p_d, p_fa, 1) == 0.0
    assert sdr.peebles(p_d, p_fa, 2) == 2.3041591328608138
    assert sdr.peebles(p_d, p_fa, 3) == 3.5677189691672773


def test_0p9():
    p_d = 0.9
    p_fa = 1e-6
    assert sdr.peebles(p_d, p_fa, 1) == 0.0
    assert sdr.peebles(p_d, p_fa, 2) == 2.717834481430985
    assert sdr.peebles(p_d, p_fa, 3) == 4.208246512218635


def test_0p999():
    p_d = 0.999
    p_fa = 1e-10
    assert sdr.peebles(p_d, p_fa, 1) == 0.0
    assert sdr.peebles(p_d, p_fa, 2) == 2.984178567081324
    assert sdr.peebles(p_d, p_fa, 3) == 4.620648951420142



"""
A module containing functions for calculating antenna gains.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.constants

from .._conversion import db
from .._helper import convert_output, export, verify_arraylike


@export
def wavelength(freq: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Calculates the wavelength $\lambda$ of an electromagnetic wave with frequency $f$.

    $$\lambda = \frac{c}{f}$$

    Arguments:
        freq: The frequency $f$ in Hz of the signal.

    Returns:
        The wavelength $\lambda$ in meters.

    Examples:
        The wavelength of a 1 GHz signal is 0.3 meters.

        .. ipython:: python

            sdr.wavelength(1e9)

    Group:
        link-budget-antennas
    """
    freq = verify_arraylike(freq, float=True, positive=True)

    lambda_ = scipy.constants.speed_of_light / freq

    return convert_output(lambda_)


@export
def parabolic_antenna(
    freq: npt.ArrayLike,
    diameter: npt.ArrayLike,
    efficiency: npt.ArrayLike = 0.55,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    r"""
    Calculates the gain $G$ and beamwidth $\theta$ of a parabolic reflector.

    Arguments:
        freq: The frequency $f$ in Hz of the signal.
        diameter: The diameter $d$ in meters of the parabolic reflector.
        efficiency: The efficiency $0 \le \eta \le 1$ of the parabolic reflector.

    Returns:
        - The gain $G$ in dBi.
        - The half-power beamwidth $\theta$ in degrees.

    Notes:
        $$G = \left( \frac{\pi d f}{c} \right)^2 \eta$$

        $$\theta = \arcsin \left( \frac{3.83 c}{\pi d f} \right)$$

    Examples:
        A 1 meter dish at 1 GHz with 55% efficiency has a gain of 17.8 dBi and a 3-dB beamwidth of 21.4 degrees.

        .. ipython:: python

            sdr.parabolic_antenna(1e9, 1)

        A 2 meter dish at 1 GHz with 55% efficiency has a gain of 23.8 dBi and a 3-dB beamwidth of 10.5 degrees.
        Since the antenna gain is proportional to the square of the reflector diameter, doubling the diameter
        results in a 6 dB increase in the gain, which we observe.

        .. ipython:: python

            sdr.parabolic_antenna(1e9, 2)

    Group:
        link-budget-antennas
    """
    freq = verify_arraylike(freq, float=True, positive=True)
    diameter = verify_arraylike(diameter, float=True, positive=True)
    efficiency = verify_arraylike(efficiency, float=True, inclusive_min=0, inclusive_max=1)

    lambda_ = wavelength(freq)  # Wavelength in meters
    G = (np.pi * diameter / lambda_) ** 2 * efficiency  # Gain in linear units
    G = db(G)  # Gain in dBi

    with np.errstate(invalid="ignore"):
        # If the argument is greater than 1, arcsin returns NaN, which is fine
        theta = np.arcsin(3.83 * lambda_ / (np.pi * diameter))  # Beamwidth in radians
    theta = np.rad2deg(theta)  # Beamwidth in degrees

    return convert_output(G), convert_output(theta)



"""
A module containing functions for calculating the capacity of a channel.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.integrate
import scipy.stats

from .._conversion import db, linear
from .._helper import convert_output, export, verify_arraylike


def Hb(x: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Computes the binary entropy function $H_b(x)$.
    """
    return scipy.stats.entropy([x, 1 - x], base=2)


@export
def bsc_capacity(p: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Calculates the capacity of a binary symmetric channel (BSC).

    Arguments:
        p: The transition probability $p$ of the BSC channel.

    Returns:
        The capacity $C$ of the channel in bits/channel use.

    Notes:
        The inputs to the BSC are $x_i \in \{0, 1\}$ and the outputs are $y_i \in \{0, 1\}$.
        The capacity of the BSC is

        $$C = 1 - H_b(p) \ \ \text{bits/channel use} .$$

    Examples:
        When the probability $p$ of bit error is 0, the capacity of the channel is 1 bit/channel use.
        However, as the probability of bit error approaches 0.5, the capacity of the channel approaches
        0.

        .. ipython:: python

            p = np.linspace(0, 1, 100); \
            C = sdr.bsc_capacity(p)

            @savefig sdr_bsc_capacity_1.png
            plt.figure(); \
            plt.plot(p, C); \
            plt.xlabel("Transition probability, $p$"); \
            plt.ylabel("Capacity (bits/channel use), $C$"); \
            plt.title("Capacity of the Binary Symmetric Channel");

    Group:
        link-budget-channel-capacity
    """
    p = verify_arraylike(p, float=True, inclusive_min=0, inclusive_max=1)

    C = 1 - Hb(p)

    return convert_output(C)


@export
def bec_capacity(p: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Calculates the capacity of a binary erasure channel (BEC).

    Arguments:
        p: The erasure probability $p$ of the BEC channel.

    Returns:
        The capacity $C$ of the channel in bits/channel use.

    Notes:
        The inputs to the BEC are $x_i \in \{0, 1\}$ and the outputs are $y_i \in \{0, 1, e\}$.
        Erasures $e$ are represented by -1. The capacity of the BEC is

        $$C = 1 - p \ \ \text{bits/channel use} .$$

    Examples:
        When the probability $p$ of erasure is 0, the capacity of the channel is 1 bit/channel use.
        However, as the probability of erasure approaches 1, the capacity of the channel approaches
        0.

        .. ipython:: python

            p = np.linspace(0, 1, 100); \
            C = sdr.bec_capacity(p)

            @savefig sdr_bec_capacity_1.png
            plt.figure(); \
            plt.plot(p, C); \
            plt.xlabel("Erasure probability, $p$"); \
            plt.ylabel("Capacity (bits/channel use), $C$"); \
            plt.title("Capacity of the Binary Erasure Channel");

    Group:
        link-budget-channel-capacity
    """
    p = verify_arraylike(p, float=True, inclusive_min=0, inclusive_max=1)

    C = 1 - p

    return convert_output(C)


@export
def awgn_capacity(snr: npt.ArrayLike, bandwidth: float | None = None) -> npt.NDArray[np.float64]:
    r"""
    Calculates the capacity of an additive white Gaussian noise (AWGN) channel.

    Arguments:
        snr: The signal-to-noise ratio $S / N$ in dB of the channel.
        bandwidth: The bandwidth $B$ of the channel in Hz. If specified, the capacity is calculated in bits/s.
            If `None`, the capacity is calculated in bits/2D.

    Returns:
        The capacity $C$ of the channel in bits/2D, or bits/s if bandwidth was specified.

    Notes:
        The inputs to the AWGN channel are $x_i \in \mathbb{C}$ and the outputs are $y_i \in \mathbb{C}$.
        The capacity of the AWGN channel is

        $$C = \log_2\left(1 + \frac{S}{N}\right) \ \ \text{bits/2D} ,$$

        where $S = \frac{1}{N} \sum_{i=0}^{N-1} \left| x_i \right|^2$ is the average signal power
        and $N = \sigma^2$ is the complex noise power. The units are bits/2D, which is equivalent to
        bits per complex channel use.

        If the channel bandwidth $B$ is specified, the channel capacity is

        $$C = B\log_2\left(1 + \frac{S}{N}\right) \ \ \text{bits/s} .$$

    Examples:
        Plot the AWGN channel capacity as a function of $S/N$. When the capacity is less than 2 bits/2D, capacity
        is in the power-limited regime. In the power-limited regime, the capacity increases linearly with signal power
        as in independent of bandwidth.

        When the capacity is greater than 2 bits/2D, capacity is in the bandwidth-limited regime. In the
        bandwidth-limited regime, the capacity increases linearly with bandwidth and logarithmically with signal power.

        .. ipython:: python

            snr = np.linspace(-30, 60, 101); \
            C = sdr.awgn_capacity(snr)

            @savefig sdr_awgn_capacity_1.png
            plt.figure(); \
            plt.semilogy(snr, C); \
            plt.axvline(sdr.shannon_limit_snr(2), color='k', linestyle='--'); \
            plt.annotate("Power-limited regime", (sdr.shannon_limit_snr(1e-2), 1e-2), xytext=(0, -20), textcoords="offset pixels", rotation=44); \
            plt.annotate("Bandwidth-limited regime", (sdr.shannon_limit_snr(8), 8), xytext=(0, -20), textcoords="offset pixels", rotation=7); \
            plt.xlabel("Signal-to-noise ratio (dB), $S/N$"); \
            plt.ylabel("Capacity (bits/2D), $C$"); \
            plt.title("Capacity of the AWGN Channel");

    Group:
        link-budget-channel-capacity
    """
    snr = verify_arraylike(snr, float=True)

    snr_linear = linear(snr)
    C = np.log2(1 + snr_linear)  # bits/2D

    if bandwidth is not None:
        bandwidth = verify_arraylike(bandwidth, float=True, positive=True)
        C *= bandwidth  # bits/s

    return convert_output(C)


@export
def biawgn_capacity(snr: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Calculates the capacity of a binary-input additive white Gaussian noise (BI-AWGN) channel.

    Arguments:
        snr: The signal-to-noise ratio $S / N = A^2 / \sigma^2$ at the output of the channel in dB.

            .. note::
                This SNR is for a real signal. In the case of soft-decision BPSK, the SNR after coherent detection is
                $S / N = 2 E_s / N_0$, where $E_s$ is the energy per symbol and $N_0$ is the noise power spectral
                density.

    Returns:
        The capacity $C$ of the channel in bits/1D.

    Notes:
        The BI-AWGN channel is defined as

        $$Y = A \cdot X + W ,$$

        where $X$ is the input with $x_i \in \{-1, 1\}$, $A$ is the signal amplitude, $W \sim \mathcal{N}(0, \sigma^2)$
        is the AWGN noise, the SNR is $A^2 / \sigma^2$, and $Y$ is the output with $y_i \in \mathbb{R}$.

        The capacity of the BI-AWGN channel is

        $$C = \max_{f_X} I(X; Y)$$

        $$I(X; Y) = H(Y) - H(Y | X) ,$$

        where $I(X; Y)$ is the mutual information between the input $X$ and output $Y$, $H(Y)$ is the differential
        entropy of the output $Y$, and $H(Y | X)$ is the conditional entropy of the output $Y$ given the input $X$.
        The maximum mutual information is achieved when $X$ is equally likely to be $-1$ or $1$.

        The conditional probability density function (PDF) of $Y$ given $X = x$ is

        $$f_{Y|X}(y|x) = \frac{1}{\sqrt{2\pi\sigma^2}} \exp \left( -\frac{(y - A x)^2}{2\sigma^2} \right) .$$

        The marginal PDF of $Y$ is

        $$f_Y(y) = \frac{1}{2} f_{Y|X}(y|1) + \frac{1}{2} f_{Y|X}(y|-1) .$$

        The differential entropy of the output $Y$ is

        $$H(Y) = - \int_{-\infty}^{\infty} f_Y(y) \log f_Y(y) \, dy .$$

        The conditional entropy of the output $Y$ given the input $X$ is

        $$H(Y | X) = - \int_{-\infty}^{\infty} \int_{-\infty}^{\infty} f_{Y|X}(y|x) \log f_{Y|X}(y|x) \, dy \, dx .$$

        In this case, since $Y = X + W$, the conditional entropy simplifies to

        $$H(Y | X) = H(W) = \frac{1}{2} \log(2\pi e \sigma^2) .$$

        The capacity of the BI-AWGN channel is

        $$
        \begin{align*}
        C
        &= H(Y) - H(Y | X) \\
        &= - \int_{-\infty}^{\infty} f_Y(y) \log f_Y(y) \, dy - \frac{1}{2} \log(2\pi e \sigma^2) .
        \end{align*}
        $$

        However, the integral must be solved using numerical methods. A convenient closed-form upper bound,
        provided by Yang, is

        $$C \leq C_{ub} = \log(2) - \log(1 + e^{-\text{SNR}}) ,$$

        where $\text{SNR} = A^2 / \sigma^2$ in linear units.

    References:
        - `Giuseppe Durisi, Chapter 3: The binary-input AWGN channel.
          <https://gdurisi.github.io/fbl-notes/bi-awgn.html>`_
        - `Pei Yang, A simple upper bound on the capacity of BI-AWGN channel.
          <https://www.jstage.jst.go.jp/article/comex/6/8/6_2017XBL0074/_pdf/-char/en>`_
        - `Tomáš Filler, Binary Additive White-Gaussian-Noise Channel.
          <https://dde.binghamton.edu/filler/mct/lectures/25/mct-lect25-bawgnc.pdf>`_

    Examples:
        .. ipython:: python

            snr = np.linspace(-30, 20, 51)

            C_ub = np.log2(2) - np.log2(1 + np.exp(-sdr.linear(snr)))
            C = sdr.biawgn_capacity(snr)

            @savefig sdr_biawgn_capacity_1.png
            plt.figure(); \
            plt.plot(snr, C_ub, linestyle="--", label="$C_{ub}$"); \
            plt.plot(snr, C, label="$C$"); \
            plt.legend(); \
            plt.xlabel("Signal-to-noise ratio (dB), $A^2/\sigma^2$"); \
            plt.ylabel("Capacity (bits/1D), $C$"); \
            plt.title("Capacity of the BI-AWGN Channel");

        .. ipython:: python

            snr = np.linspace(-30, 20, 51)
            p = sdr.Q(np.sqrt(sdr.linear(snr)))
            C_hard = sdr.bsc_capacity(p)
            C_soft = sdr.biawgn_capacity(snr)

            @savefig sdr_biawgn_capacity_2.png
            plt.figure(); \
            plt.plot(snr, C_hard, label="BSC (hard)"); \
            plt.plot(snr, C_soft, label="BI-AWGN (soft)"); \
            plt.legend(); \
            plt.xlabel("Signal-to-noise ratio (dB), $A^2/\sigma^2$"); \
            plt.ylabel("Capacity (bits/1D), $C$"); \
            plt.title("Capacity of the BSC (hard-decision) and BI-AWGN (soft-decision) Channels");

    Group:
        link-budget-channel-capacity
    """
    snr = verify_arraylike(snr, float=True)

    @np.vectorize
    def _calculate(snr: float) -> float:
        sigma2 = 1  # Noise power (variance), sigma^2
        A2 = linear(snr) * sigma2  # Signal power, A^2

        # f_Y|X(y | 1) is the PDF of Y given X = 1 was sent
        f_y_p1 = scipy.stats.norm(np.sqrt(A2), np.sqrt(sigma2)).pdf

        # f_Y|X(y | -1) is the PDF of Y given X = -1 was sent
        f_y_n1 = scipy.stats.norm(-np.sqrt(A2), np.sqrt(sigma2)).pdf

        # f_Y(y) is the marginal PDF of Y. This assumes equal probability of X = 0 and X = 1, which is required to
        # maximize the mutual information I(X; Y) and achieve capacity.
        def f_y(y):
            return 0.5 * f_y_p1(y) + 0.5 * f_y_n1(y)

        # H(Y | X) is the conditional entropy of the output Y given the input X. Since Y = X + W, the conditional
        # entropy is the differential entropy of the noise W.
        H_y_x = 0.5 * np.log2(2 * np.pi * np.e * sigma2)

        # H(Y) is the differential entropy of the output Y
        H_y = scipy.integrate.quad(lambda y: np.nan_to_num(-f_y(y) * np.log2(f_y(y))), -np.inf, np.inf)[0]

        # I(X; Y) is the mutual information between the input X and the output Y. This is the maximum achievable
        # mutual information and the channel capacity.
        I_x_y = H_y - H_y_x

        return I_x_y  # bits/1D

    with np.errstate(divide="ignore", invalid="ignore"):
        rho = _calculate(snr)

    return convert_output(rho)


@export
def shannon_limit_ebn0(rho: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Calculates the Shannon limit on the bit energy-to-noise power spectral density ratio $E_b/N_0$ in the AWGN
    channel.

    Arguments:
        rho: The nominal spectral efficiency $\rho$ of the modulation in bits/2D.

    Returns:
        The Shannon limit on $E_b/N_0$ in dB.

    See Also:
        sdr.awgn_capacity

    Notes:
        $$C = \rho = \log_2\left(1 + \frac{S}{N}\right) = \log_2\left(1 + \rho\frac{E_b}{N_0}\right) \ \ \text{bits/2D}$$
        $$\frac{E_b}{N_0} = \frac{2^{\rho} - 1}{\rho}$$

    Examples:
        The *absolute Shannon limit* on power efficiency is -1.59 dB $E_b/N_0$. This can only be achieved when the
        channel capacity approaches 0.

        .. ipython:: python

            sdr.shannon_limit_ebn0(0)

        The Shannon limit for various FEC code rates is shown below. The modulation is assumed to be binary.

        .. ipython:: python

            rate = np.array([1/8, 1/4, 1/3, 1/2, 2/3, 3/4, 7/8])  # Code rate, n/k
            rho = 2 * rate  # bits/2D
            sdr.shannon_limit_ebn0(rho)

        Plot the AWGN channel capacity as a function of $E_b/N_0$.

        .. ipython:: python

            rho = np.logspace(-2, 1, 101)
            ebn0 = sdr.shannon_limit_ebn0(rho)

            @savefig sdr_shannon_limit_ebn0_1.png
            plt.figure(); \
            plt.semilogy(ebn0, rho); \
            plt.xlabel("Bit energy-to-noise PSD ratio (dB), $E_b/N_0$"); \
            plt.ylabel("Capacity (bits/2D), $C$"); \
            plt.title("Capacity of the AWGN Channel");

    Group:
        link-budget-channel-capacity
    """
    rho = verify_arraylike(rho, float=True, non_negative=True)

    @np.vectorize
    def _calculate(rho: float) -> float:
        if rho == 0:
            ebn0 = np.log(2)
        else:
            rho = float(rho)  # If rho is an integer, the below equation can error for large rho
            ebn0 = (2**rho - 1) / rho

        return db(ebn0)

    ebn0 = _calculate(rho)

    return convert_output(ebn0)


@export
def shannon_limit_snr(rho: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Calculates the Shannon limit on the signal-to-noise ratio $S/N$ in the AWGN channel.

    Arguments:
        rho: The nominal spectral efficiency $\rho$ of the modulation in bits/2D.

    Returns:
        The Shannon limit on $S/N$ in dB.

    See Also:
        sdr.awgn_capacity

    Notes:
        $$C = \rho = \log_2\left(1 + \frac{S}{N}\right) \ \ \text{bits/2D}$$
        $$\frac{S}{N} = 2^{\rho} - 1$$

    Examples:
        Plot the AWGN channel capacity as a function of $S/N$. When the capacity is less than 2 bits/2D, capacity
        is in the power-limited regime. In the power-limited regime, the capacity increases linearly with signal power
        as in independent of bandwidth.

        When the capacity is greater than 2 bits/2D, capacity is in the bandwidth-limited regime. In the
        bandwidth-limited regime, the capacity increases linearly with bandwidth and logarithmically with signal power.

        .. ipython:: python

            snr = np.linspace(-30, 60, 101); \
            C = sdr.awgn_capacity(snr)

            @savefig sdr_shannon_limit_snr_1.png
            plt.figure(); \
            plt.semilogy(snr, C); \
            plt.axvline(sdr.shannon_limit_snr(2), color='k', linestyle='--'); \
            plt.annotate("Power-limited regime", (sdr.shannon_limit_snr(1e-2), 1e-2), xytext=(0, -20), textcoords="offset pixels", rotation=44); \
            plt.annotate("Bandwidth-limited regime", (sdr.shannon_limit_snr(8), 8), xytext=(0, -20), textcoords="offset pixels", rotation=7); \
            plt.xlabel("Signal-to-noise ratio (dB), $S/N$"); \
            plt.ylabel("Capacity (bits/2D), $C$"); \
            plt.title("Capacity of the AWGN Channel");

    Group:
        link-budget-channel-capacity
    """
    rho = verify_arraylike(rho, float=True, non_negative=True)

    @np.vectorize
    def _calculate(rho: float) -> float:
        rho = float(rho)  # If rho is an integer, the below equation can error for large rho
        return db(2**rho - 1)

    snr = _calculate(rho)

    return convert_output(snr)
